from __future__ import annotations

from typing import Any, cast

from coredis._utils import EncodingInsensitiveDict
from coredis.response._callbacks import (
    ClusterMergeMapping,
    ResponseCallback,
)
from coredis.response._utils import flat_pairs_to_dict
from coredis.typing import (
    AnyStr,
    ResponsePrimitive,
    ResponseType,
    StringT,
)


class SampleCallback(
    ResponseCallback[
        list[int | float],
        tuple[int, float] | tuple[()],
    ]
):
    def transform(
        self,
        response: list[int | float],
    ) -> tuple[int, float] | tuple[()]:
        return (int(response[0]), float(response[1])) if response else ()


class SamplesCallback(
    ResponseCallback[
        list[list[int | float]] | None,
        tuple[tuple[int, float], ...] | tuple[()],
    ],
):
    def transform(
        self,
        response: list[list[int | float]] | None,
    ) -> tuple[tuple[int, float], ...] | tuple[()]:
        if response:
            return tuple(cast(tuple[int, float], SampleCallback().transform(r)) for r in response)
        return ()


class TimeSeriesInfoCallback(
    ResponseCallback[dict[ResponsePrimitive, ResponseType], dict[AnyStr, ResponseType]]
):
    def transform(
        self,
        response: dict[ResponsePrimitive, ResponseType],
    ) -> dict[AnyStr, ResponseType]:
        dct = EncodingInsensitiveDict(response)
        if "labels" in dct:
            dct["labels"] = dict(dct["labels"])
        if "Chunks" in dct:
            dct["Chunks"] = [flat_pairs_to_dict(chunk) for chunk in dct["Chunks"]]
        return dict(dct)


class TimeSeriesCallback(
    ResponseCallback[
        dict[StringT, list[dict[StringT, StringT] | list[int | float]]],
        dict[AnyStr, tuple[dict[AnyStr, AnyStr], tuple[int, float] | tuple[()]]],
    ]
):
    def transform(
        self,
        response: dict[StringT, list[dict[StringT, StringT] | list[int | float]]],
    ) -> dict[AnyStr, tuple[dict[AnyStr, AnyStr], tuple[int, float] | tuple[()]]]:
        return {
            cast(AnyStr, k): (
                cast(dict[AnyStr, AnyStr], v[0]),
                cast(tuple[int, float], tuple(v[1])),
            )
            for k, v in response.items()
        }


class TimeSeriesMultiCallback(
    ResponseCallback[
        dict[StringT, list[dict[StringT, StringT | list[StringT]] | list[list[int | float]]]],
        dict[
            AnyStr,
            tuple[dict[AnyStr, AnyStr], tuple[tuple[int, float], ...] | tuple[()]],
        ],
    ]
):
    def transform(
        self,
        response: dict[
            StringT, list[dict[StringT, StringT | list[StringT]] | list[list[int | float]]]
        ],
    ) -> dict[
        AnyStr,
        tuple[dict[AnyStr, AnyStr], tuple[tuple[int, float], ...] | tuple[()]],
    ]:
        return {
            cast(AnyStr, k): (
                cast(dict[AnyStr, AnyStr], r[0]),
                tuple(
                    cast(tuple[int, float], SampleCallback().transform(cast(list[int | float], t)))
                    for t in r[-1]
                ),
            )
            for k, r in response.items()
        }


class ClusterMergeTimeSeries(ClusterMergeMapping[AnyStr, tuple[Any, ...]]):
    def __init__(self) -> None:
        super().__init__(value_combine=self.merge)

    def merge(self, values: Any) -> tuple[dict[AnyStr, AnyStr], tuple[tuple[int, float], ...]]:
        merged_labels: dict[AnyStr, AnyStr] = {}
        merged_series: tuple[tuple[int, float], ...] = ()
        for value in values:
            merged_labels.update(value[0])
            merged_series = merged_series + value[1]
        return merged_labels, tuple(merged_series)



from __future__ import annotations

from collections.abc import Set

from coredis.typing import (
    Literal,
    Mapping,
    NamedTuple,
    OrderedDict,
    RedisValueT,
    StringT,
    TypedDict,
)

#: Response from `CLIENT INFO <https://redis.io/commands/client-info>`__
#:
#: - ``id``: a unique 64-bit client ID
#: - ``addr``: address/port of the client
#: - ``laddr``: address/port of local address client connected to (bind address)
#: - ``fd``: file descriptor corresponding to the socket
#: - ``name``: the name set by the client with CLIENT SETNAME
#: - ``age``: total duration of the connection in seconds
#: - ``idle``: idle time of the connection in seconds
#: - ``flags``: client flags
#: - ``db``: current database ID
#: - ``sub``: number of channel subscriptions
#: - ``psub``: number of pattern matching subscriptions
#: - ``multi``: number of commands in a MULTI/EXEC context
#: - ``qbuf``: query buffer length (0 means no query pending)
#: - ``qbuf-free``: free space of the query buffer (0 means the buffer is full)
#: - ``argv-mem``: incomplete arguments for the next command (already extracted from query buffer)
#: - ``multi-mem``: memory is used up by buffered multi commands. Added in Redis 7.0
#: - ``obl``: output buffer length
#: - ``oll``: output list length (replies are queued in this list when the buffer is full)
#: - ``omem``: output buffer memory usage
#: - ``tot-mem``: total memory consumed by this client in its various buffers
#: - ``events``: file descriptor events
#: - ``cmd``: last command played
#: - ``user``: the authenticated username of the client
#: - ``redir``: client id of current client tracking redirection
#: - ``resp``: client RESP protocol version. Added in Redis 7.0
ClientInfo = TypedDict(
    "ClientInfo",
    {
        "id": int,
        "addr": str,
        "laddr": str,
        "fd": int,
        "name": str,
        "age": int,
        "idle": int,
        "flags": str,
        "db": int,
        "sub": int,
        "psub": int,
        "multi": int,
        "qbuf": int,
        "qbuf-free": int,
        "argv-mem": int,
        "multi-mem": int,
        "obl": int,
        "oll": int,
        "omem": int,
        "tot-mem": int,
        "events": str,
        "cmd": str,
        "user": str,
        "redir": int,
        "resp": str,
    },
)

#: Script/Function flags
#: See: `<https://redis.io/topics/lua-api#a-namescriptflagsa-script-flags>`__
ScriptFlag = Literal[
    "no-writes",
    "allow-oom",
    "allow-stale",
    "no-cluster",
    b"no-writes",
    b"allow-oom",
    b"allow-stale",
    b"no-cluster",
]


class FunctionDefinition(TypedDict):
    """
    Function definition as returned by `FUNCTION LIST <https://redis.io/commands/function-list>`__
    """

    #: the name of the function
    name: StringT
    #: the description of the function
    description: StringT
    #: function flags
    flags: set[ScriptFlag]


class LibraryDefinition(TypedDict):
    """
    Library definition as returned by `FUNCTION LIST <https://redis.io/commands/function-list>`__
    """

    #: the name of the library
    name: StringT
    #: the engine used by the library
    engine: Literal["LUA"]
    #: Mapping of function names to functions in the library
    functions: dict[StringT, FunctionDefinition]
    #: The library's source code
    library_code: StringT | None


class ScoredMember(NamedTuple):
    """
    Member of a sorted set
    """

    #: The sorted set member name
    member: StringT
    #: Score of the member
    score: float


class GeoCoordinates(NamedTuple):
    """
    A longitude/latitude pair identifying a location
    """

    #: Longitude
    longitude: float
    #: Latitude
    latitude: float


ScoredMembers = tuple[ScoredMember, ...]


class GeoSearchResult(NamedTuple):
    """
    Structure of a geo query
    """

    #: Place name
    name: StringT
    #: Distance
    distance: float | None
    #: GeoHash
    geohash: int | None
    #: Lat/Lon
    coordinates: GeoCoordinates | None


#: Definition of a redis command
#: See: `<https://redis.io/topics/key-specs>`__
#:
#: - ``name``: This is the command's name in lowercase
#: - ``arity``: Arity is the number of arguments a command expects.
#: - ``flags``: See `<https://redis.io/commands/command#flags>`__
#: - ``first-key``: This value identifies the position of the command's first key name argumen
#: - ``last-key``: This value identifies the position of the command's last key name argument
#: - ``step``: This value is the step, or increment, between the first key and last key values
#:   where the keys are.
#: - ``acl-categories``: This is an array of simple strings that are the ACL categories to which
#:   the command belongs
#: - ``tips``: Helpful information about the command
#: - ``key-specification``: This is an array consisting of the command's key specifications
#: - ``sub-commands``: This is an array containing all of the command's subcommands, if any
Command = TypedDict(
    "Command",
    {
        "name": str,
        "arity": int,
        "flags": Set[str],
        "first-key": int,
        "last-key": int,
        "step": int,
        "acl-categories": Set[str] | None,
        "tips": Set[str] | None,
        "key-specifications": Set[Mapping[str, int | str | Mapping]] | None,  # type: ignore
        "sub-commands": tuple[str, ...] | None,
    },
)


class RoleInfo(NamedTuple):
    """
    Redis instance role information
    """

    #:
    role: str
    #:
    offset: int | None = None
    #:
    status: str | None = None
    #:
    slaves: tuple[dict[str, str | int], ...] | None = None
    #:
    masters: tuple[str, ...] | None = None


class StreamEntry(NamedTuple):
    """
    Structure representing an entry in a redis stream
    """

    identifier: StringT
    field_values: OrderedDict[StringT, StringT]


#: Details of a stream
#: See: `<https://redis.io/commands/xinfo-stream>`__
StreamInfo = TypedDict(
    "StreamInfo",
    {
        "first-entry": StreamEntry | None,
        "last-entry": StreamEntry | None,
        "length": int,
        "radix-tree-keys": int,
        "radix-tree-nodes": int,
        "groups": int | list[dict],  # type: ignore
        "last-generated-id": str,
        "max-deleted-entry-id": str,
        "recorded-first-entry-id": str,
        "entries-added": int,
        "entries-read": int,
        "entries": tuple[StreamEntry, ...] | None,
    },
)


class StreamPending(NamedTuple):
    """
    Summary response from
    `XPENDING <https://redis.io/commands/xpending#summary-form-of-xpending>`__
    """

    pending: int
    minimum_identifier: StringT
    maximum_identifier: StringT
    consumers: OrderedDict[StringT, int]


class StreamPendingExt(NamedTuple):
    """
    Extended form response from
    `XPENDING <https://redis.io/commands/xpending#extended-form-of-xpending>`__
    """

    identifier: StringT
    consumer: StringT
    idle: int
    delivered: int


#: Response from `SLOWLOG GET <https://redis.io/commands/slowlog-get>`__
class SlowLogInfo(NamedTuple):
    #: A unique progressive identifier for every slow log entry.
    id: int
    #: The unix timestamp at which the logged command was processed.
    start_time: int
    #: The amount of time needed for its execution, in microseconds.
    duration: int
    #: The array composing the arguments of the command.
    command: list[StringT]
    #: Client IP address:port
    client_addr: StringT
    #: Client name
    client_name: StringT


class LCSMatch(NamedTuple):
    """
    An instance of an LCS match
    """

    #: Start/end offset of the first string
    first: tuple[int, int]
    #: Start/end offset of the second string
    second: tuple[int, int]
    #: Length of the match
    length: int | None


class LCSResult(NamedTuple):
    """
    Results from `LCS <https://redis.io/commands/lcs>`__
    """

    #: matches
    matches: tuple[LCSMatch, ...]
    #: Length of longest match
    length: int


class ClusterNode(TypedDict):
    host: str
    port: int
    node_id: str | None
    server_type: Literal["master", "slave"] | None


class ClusterNodeDetail(TypedDict):
    id: str
    flags: tuple[str, ...]
    host: str
    port: int
    master: str | None
    ping_sent: int
    pong_recv: int
    link_state: str
    slots: list[int]
    migrations: list[dict[str, RedisValueT]]


class PubSubMessage(TypedDict):
    #: One of the following:
    #:
    #: subscribe
    #:   Server response when a client subscribes to a channel(s)
    #: unsubscribe
    #:   Server response when a client unsubscribes from a channel(s)
    #: psubscribe
    #:   Server response when a client subscribes to a pattern(s)
    #: punsubscribe
    #:   Server response when a client unsubscribes from a pattern(s)
    #: ssubscribe
    #:   Server response when a client subscribes to a shard channel(s)
    #: sunsubscribe
    #:   Server response when a client unsubscribes from a shard channel(s)
    #: message
    #:   A message received from subscribing to a channel
    #: pmessage
    #:   A message received from subscribing to a pattern
    type: str
    #: The channel subscribed to or unsubscribed from or the channel a message was published to
    channel: StringT
    #: The pattern that was subscribed to or unsubscribed from or to which a received message was
    #: routed to
    pattern: StringT | None
    #: - If ``type`` is one of ``{message, pmessage}`` this is the actual published message
    #: - If ``type`` is one of
    #:   ``{subscribe, psubscribe, ssubscribe, unsubscribe, punsubscribe, sunsubscribe}``
    #:   this will be an :class:`int` corresponding to the  number of channels and patterns that the
    #:   connection is currently subscribed to.
    data: int | StringT


class VectorData(TypedDict):
    #: The quantization type as a string (``fp32``, ``bin`` or ``q8``)
    quantization: str
    #: Raw bytes representation of the vector
    blob: bytes
    #: The L2 norm of the vector before normalization
    l2_norm: float
    #: If the vector is quantized as q8, the quantization range
    quantization_range: float | None



from __future__ import annotations

import dataclasses
import functools
import textwrap
import warnings
from typing import Any

from packaging import version

from coredis.globals import (
    CACHEABLE_COMMANDS,
    COMMAND_FLAGS,
    READONLY_COMMANDS,
    ROUTING_STRATEGIES,
)
from coredis.typing import (
    Callable,
    NamedTuple,
    P,
    R,
    add_runtime_checks,
)

from ._routing import RoutingStrategy
from ._utils import check_version, redis_command_link
from .constants import CommandFlag, CommandGroup, CommandName, NodeFlag
from .request import CommandRequest


class RedirectUsage(NamedTuple):
    reason: str
    warn: bool


@dataclasses.dataclass
class CommandDetails:
    command: bytes
    group: CommandGroup | None
    version_introduced: version.Version | None
    version_deprecated: version.Version | None
    _arguments: dict[str, dict[str, str]] | None
    cluster: ClusterCommandConfig
    flags: set[CommandFlag]
    redirect_usage: RedirectUsage | None
    arguments: dict[str, version.Version] = dataclasses.field(
        init=False, default_factory=lambda: {}
    )

    def __post_init__(self) -> None:
        self.arguments = {
            k: version.Version(v["version_introduced"])
            for k, v in (self._arguments or {}).items()
            if v.get("version_introduced")
        }


@dataclasses.dataclass
class ClusterCommandConfig:
    enabled: bool = True
    routing_strategy: RoutingStrategy[Any] | None = None

    @property
    def multi_node(self) -> bool:
        flag = self.routing_strategy.route if self.routing_strategy else None
        return flag in [
            NodeFlag.ALL,
            NodeFlag.PRIMARIES,
            NodeFlag.REPLICAS,
            NodeFlag.SLOT_ID,
        ]


def redis_command(
    command_name: CommandName,
    *,
    group: CommandGroup | None = None,
    version_introduced: str | None = None,
    version_deprecated: str | None = None,
    deprecation_reason: str | None = None,
    redirect_usage: RedirectUsage | None = None,
    arguments: dict[str, dict[str, str]] | None = None,
    flags: set[CommandFlag] | None = None,
    cluster: ClusterCommandConfig = ClusterCommandConfig(),
    cacheable: bool | None = None,
) -> Callable[[Callable[P, CommandRequest[R]]], Callable[P, CommandRequest[R]]]:
    readonly = False
    if flags and CommandFlag.READONLY in flags:
        READONLY_COMMANDS.add(command_name)
        readonly = True

    COMMAND_FLAGS[command_name] = flags or set()

    if not readonly and cacheable:  # noqa
        raise RuntimeError(f"Can't decorate non readonly command {command_name} with cache config")
    if cacheable:
        CACHEABLE_COMMANDS.add(command_name)

    if cluster.routing_strategy:
        ROUTING_STRATEGIES[command_name] = cluster.routing_strategy
    # elif cluster.route == NodeFlag.SLOT_ID:
    #    ROUTING_STRATEGIES[command_name] = SlotRangeStrategy(cluster.combine)
    # elif cluster.combine and cluster.route:
    #    ROUTING_STRATEGIES[command_name] = FanoutStrategy(cluster.route, cluster.combine)
    # elif cluster.route == NodeFlag.RANDOM:
    #    ROUTING_STRATEGIES[command_name] = RandomStrategy()
    command_details = CommandDetails(
        command_name,
        group,
        version.Version(version_introduced) if version_introduced else None,
        version.Version(version_deprecated) if version_deprecated else None,
        arguments,
        cluster or ClusterCommandConfig(),
        flags or set(),
        redirect_usage,
    )

    def wrapper(
        func: Callable[P, CommandRequest[R]],
    ) -> Callable[P, CommandRequest[R]]:
        runtime_checkable = add_runtime_checks(func)

        @functools.wraps(func)
        def wrapped(*args: P.args, **kwargs: P.kwargs) -> CommandRequest[R]:
            from coredis import Redis, RedisCluster

            is_regular_client = isinstance(args[0], (Redis, RedisCluster))
            if redirect_usage and is_regular_client:
                if redirect_usage.warn:
                    warnings.warn(redirect_usage.reason, UserWarning, stacklevel=2)
                else:
                    raise NotImplementedError(redirect_usage.reason)
            runtime_checking = not getattr(args[0], "noreply", None) and is_regular_client
            check_version(
                args[0],  # type: ignore
                func.__name__,
                command_details,
                deprecation_reason,
                kwargs,
            )
            return (func if not runtime_checking else runtime_checkable)(*args, **kwargs)

        wrapped.__doc__ = textwrap.dedent(wrapped.__doc__ or "")
        if group:
            wrapped.__doc__ = f"""
{wrapped.__doc__}

Redis command documentation: {redis_command_link(command_name)}
"""
        if version_introduced or command_details.arguments:
            wrapped.__doc__ += """
Compatibility:
"""

        if version_introduced:
            wrapped.__doc__ += f"""
- New in :redis-version:`{version_introduced}`
"""
        if version_deprecated and deprecation_reason:
            wrapped.__doc__ += f"""
.. caution:: Deprecated in :redis-version:`{version_deprecated}`
   {deprecation_reason.strip()}
"""
        elif version_deprecated:
            wrapped.__doc__ += f"""
.. caution:: Deprecated in :redis-version:`{version_deprecated}`
"""
        if command_details.arguments:
            for argument, min_version in command_details.arguments.items():
                wrapped.__doc__ += f"""
- :paramref:`{argument}`: New in :redis-version:`{min_version}`
"""
        if cacheable:
            wrapped.__doc__ += """
.. hint:: Supports client side caching
"""
        if redirect_usage:
            if redirect_usage.warn:
                preamble = f".. warning:: Using ``{func.__name__}`` directly is not recommended."
            else:
                preamble = f".. danger:: Using ``{func.__name__}`` directly is not supported."
            wrapped.__doc__ += f"""
{preamble} {redirect_usage.reason}
"""

        setattr(wrapped, "__coredis_command", command_details)
        return wrapped

    return wrapper



from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING, Any

from coredis.connection import ClusterConnection, TCPLocation
from coredis.typing import Literal, Unpack

if TYPE_CHECKING:
    from coredis import Redis
    from coredis.pool import ConnectionPoolParams


@dataclasses.dataclass(unsafe_hash=True)
class ClusterNodeLocation(TCPLocation):
    """
    Represents a cluster node (primary or replica) in a redis cluster
    """

    server_type: Literal["primary", "replica"] | None = None
    node_id: str | None = None
    priority: int = 100

    @property
    def name(self) -> str:
        return f"{self.host}:{self.port}"

    @property
    def is_primary(self) -> bool:
        return self.server_type == "primary"

    def as_client(
        self, **client_args: Unpack[ConnectionPoolParams[ClusterConnection]]
    ) -> Redis[Any]:
        from coredis import Redis
        from coredis.pool import ConnectionPool

        pool = ConnectionPool[ClusterConnection](
            connection_class=ClusterConnection,
            location=TCPLocation(self.host, self.port),
            decode_responses=client_args.get("decode_responses", False),
            encoding=client_args.get("encoding", "utf-8"),
            username=client_args.get("username"),
            password=client_args.get("password"),
            credential_provider=client_args.get("credential_provider"),
            ssl_context=client_args.get("ssl_context"),
            stream_timeout=client_args.get("stream_timeout"),
            connect_timeout=client_args.get("connect_timeout"),
        )
        return Redis(connection_pool=pool)



from __future__ import annotations

from coredis._enum import CaseAndEncodingInsensitiveEnum


class PubSubMessageTypes(CaseAndEncodingInsensitiveEnum):
    MESSAGE = b"message"
    PMESSAGE = b"pmessage"
    SMESSAGE = b"smessage"
    SUBSCRIBE = b"subscribe"
    UNSUBSCRIBE = b"unsubscribe"
    PSUBSCRIBE = b"psubscribe"
    PUNSUBSCRIBE = b"punsubscribe"
    SSUBSCRIBE = b"ssubscribe"
    SUNSUBSCRIBE = b"sunsubscribe"


PUBLISH_MESSAGE_TYPES = {
    PubSubMessageTypes.MESSAGE.value,
    PubSubMessageTypes.PMESSAGE.value,
    PubSubMessageTypes.SMESSAGE.value,
}
SUBSCRIBE_MESSAGE_TYPES = {
    PubSubMessageTypes.SUBSCRIBE.value,
    PubSubMessageTypes.PSUBSCRIBE.value,
    PubSubMessageTypes.SSUBSCRIBE.value,
}
UNSUBSCRIBE_MESSAGE_TYPES = {
    PubSubMessageTypes.UNSUBSCRIBE.value,
    PubSubMessageTypes.PUNSUBSCRIBE.value,
    PubSubMessageTypes.SUNSUBSCRIBE.value,
}

SUBUNSUB_MESSAGE_TYPES = SUBSCRIBE_MESSAGE_TYPES | UNSUBSCRIBE_MESSAGE_TYPES



from __future__ import annotations

import dataclasses

from coredis._json import json
from coredis.typing import (
    AnyStr,
    Generic,
    ResponseType,
    StringT,
)


@dataclasses.dataclass
class SearchDocument(Generic[AnyStr]):
    """
    Search document as returned by `FT.SEARCH <https://redis.io/commands/ft.search>`__
    """

    #: Document id
    id: StringT
    #: Search score if the :paramref:`~coredis.modules.search.Search.search.withscores`
    #: option was used
    score: float | None
    #: Explanation of the score if the
    #:  :paramref:`~coredis.modules.search.Search.search.explainscore` option was used
    score_explanation: list[AnyStr] | None
    #: Payload associated with the document if
    #:  :paramref:`~coredis.modules.search.Search.search.withpayloads` was used
    payload: StringT | None
    sortkeys: StringT | None
    #: Mapping of properties returned for the document
    properties: dict[AnyStr, ResponseType]


@dataclasses.dataclass
class SearchResult(Generic[AnyStr]):
    """
    Search results as returned by `FT.SEARCH <https://redis.io/commands/ft.search>`__
    """

    #: The total number of results found for the query
    total: int
    #: The documents returned by the query
    documents: tuple[SearchDocument[AnyStr], ...]


@dataclasses.dataclass
class SearchAggregationResult(Generic[AnyStr]):
    """
    Search aggregations as returned by `FT.AGGREGATE <https://redis.io/commands/ft.aggregate>`__
    """

    #: The aggregation results
    results: list[dict[StringT, ResponseType]]
    #: The cursor id if :paramref:`~coredis.modules.search.aggregate.with_cursor` was `True`
    cursor: int | None

    def __post_init__(self) -> None:
        for idx, result in enumerate(self.results):
            json_key = b"$" if b"$" in result else "$" if "$" in result else None
            if json_key:
                self.results[idx] = json.loads(result.pop(json_key))


@dataclasses.dataclass
class HybridResult(Generic[AnyStr]):
    total_results: int
    execution_time: float
    warnings: list[AnyStr]
    results: list[dict[AnyStr, AnyStr]]


@dataclasses.dataclass
class AutocompleteSuggestion(Generic[AnyStr]):
    """
    Autocomplete suggestion as returned by `FT.SUGGET <https://redis.io/commands/ft.sugget>`__
    """

    #: the suggestion string
    string: AnyStr
    #: the score of the suggestion if
    #:  :paramref:`~coredis.modules.autocomplete.Autocomplete.sugget.withscores` was used
    score: float | None
    #: the payload associated with the suggestion if
    #:  :paramref:`~coredis.modules.autocomplete.Autocomplete.sugget.withpayloads` was used
    payload: AnyStr | None



from __future__ import annotations

import dataclasses
import socket

from anyio import connect_tcp, fail_after
from anyio.abc import ByteStream, SocketAttribute

from coredis.typing import Unpack

from ._base import BaseConnection, BaseConnectionParams, Location


@dataclasses.dataclass(unsafe_hash=True)
class TCPLocation(Location):
    """Location of a redis instance listening on a tcp port"""

    #: hostname of the server
    host: str
    #: the port the server is listening on
    port: int

    def __repr__(self) -> str:
        return f"<host={self.host},port={self.port}>"

    async def check(self) -> bool:
        try:
            async with await connect_tcp(self.host, self.port):
                return True
        except OSError:
            return False
        return False


class TCPConnection(BaseConnection):
    location: TCPLocation

    def __init__(
        self,
        location: TCPLocation,
        *,
        socket_keepalive: bool | None = None,
        socket_keepalive_options: dict[int, int | bytes] | None = None,
        **kwargs: Unpack[BaseConnectionParams],
    ):
        super().__init__(location, **kwargs)
        self._socket_keepalive = socket_keepalive
        self._socket_keepalive_options: dict[int, int | bytes] = socket_keepalive_options or {}
        # FIXME: this is only for backward compatibility as 6.0 still had
        # host/port in TCP connections
        self.host = self.location.host
        self.port = self.location.port

    async def _connect(self) -> ByteStream:
        with fail_after(self._connect_timeout):
            if self._ssl_context:
                connection: ByteStream = await connect_tcp(
                    self.location.host,
                    self.location.port,
                    tls=True,
                    ssl_context=self._ssl_context,
                    tls_standard_compatible=False,
                )
            else:
                connection = await connect_tcp(self.location.host, self.location.port)
            sock = connection.extra(SocketAttribute.raw_socket, default=None)
            if sock is not None:
                if self._socket_keepalive:  # TCP_KEEPALIVE
                    sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
                    for k, v in self._socket_keepalive_options.items():
                        sock.setsockopt(socket.SOL_TCP, k, v)
            return connection

    def describe(self) -> str:
        return f"Connection<host={self.location.host},port={self.location.port},db={self._db}>"



from __future__ import annotations

import contextvars
from abc import ABC, abstractmethod
from typing import Any
from urllib.parse import parse_qs, unquote, urlparse

from anyio import (
    Lock,
    create_task_group,
)
from anyio.abc import TaskGroup

from coredis._utils import query_param_to_bool
from coredis.connection._base import (
    BaseConnectionParams,
    ConnectionT,
    Location,
    RedisSSLContext,
)
from coredis.connection._tcp import TCPLocation
from coredis.connection._uds import UnixDomainSocketLocation
from coredis.typing import (
    AsyncContextManager,
    Callable,
    ClassVar,
    Generic,
    NotRequired,
    ParamSpec,
    Self,
    TypeVar,
    Unpack,
)

BaseConnectionPoolParamsT = TypeVar(
    "BaseConnectionPoolParamsT", bound="BaseConnectionPoolParams[Any]"
)

ConnectionPoolT = TypeVar("ConnectionPoolT", bound="BaseConnectionPool[Any]")
P = ParamSpec("P")
R = TypeVar("R")


class BaseConnectionPoolParams(BaseConnectionParams, Generic[ConnectionT]):
    """
    Connection pool parameters accepted by :class:`coredis.pool.BaseConnectionPool`
    """

    #: The connection class to use when creating new connections
    connection_class: NotRequired[type[ConnectionT]]
    #: Maximum connections to grow the pool.
    #: Once the limit is reached clients will block to wait for a connection
    #: to be returned to the pool.
    max_connections: NotRequired[int | None]
    #: Number of seconds to block when trying to obtain a connection.
    timeout: NotRequired[float | None]


class BaseConnectionPool(ABC, Generic[ConnectionT]):
    URL_QUERY_ARGUMENT_PARSERS: ClassVar[
        dict[str, Callable[..., int | float | bool | str | None]]
    ] = {
        "max_connections": int,
        "timeout": int,
        "client_name": str,
        "stream_timeout": float,
        "connect_timeout": float,
        "max_idle_time": int,
        "noreply": query_param_to_bool,
        "noevict": query_param_to_bool,
        "notouch": query_param_to_bool,
        "db": int,
    }

    def __init__(
        self,
        *,
        connection_class: type[ConnectionT],
        location: Location | None = None,
        max_connections: int | None = None,
        timeout: float | None = None,
        **connection_kwargs: Unpack[BaseConnectionParams],
    ) -> None:
        """
        Base (Abstract) connection pool

        :param connection_class: The connection class to use when creating new connections
        :param max_connections: Maximum connections to grow the pool.
         Once the limit is reached clients will block to wait for a connection
         to be returned to the pool.
        :param timeout: Number of seconds to block when trying to obtain a connection.
        :param connection_kwargs: arguments to pass to the :paramref:`connection_class`
         constructor when creating a new connection
        """
        self.connection_class = connection_class
        self.connection_kwargs = connection_kwargs
        self.max_connections = max_connections or 64
        self.timeout = timeout
        self.decode_responses = bool(self.connection_kwargs.get("decode_responses", False))
        self.encoding = str(self.connection_kwargs.get("encoding", "utf-8"))
        self.connect_timeout = self.connection_kwargs.get("connect_timeout", 1)
        self.location = location
        # reference count for context manager to support this pool being re-entered.
        self._counter = 0
        # context to track whether the intializing task (anchor) is active
        self._anchor_active = contextvars.ContextVar("parent_active", default=False)
        self._anchor_reset_token: contextvars.Token[bool] | None = None
        self._initialization_lock = Lock()
        self._task_group: TaskGroup | None = None

    def __repr__(self) -> str:
        return f"{type(self).__name__}<{self.location}>"

    async def __aenter__(self) -> Self:
        async with self._initialization_lock:
            if self._counter == 0:
                self._task_group = create_task_group()
                self._anchor_reset_token = self._anchor_active.set(True)
                self._counter += 1
                await self._task_group.__aenter__()
                try:
                    await self._initialize()
                except BaseException:
                    await self.__aexit__(None, None, None)
                    raise
            else:
                if not self._anchor_active.get():
                    raise RuntimeError(
                        "Implicit concurrent connection pool sharing detected. "
                        "You must explicitly enter the pool in a parent task "
                        "before spawning concurrent tasks that "
                        "share it. (For more details see "
                        "https://coredis.readthedocs.io/en/stable/handbook/connections.html#sharing-a-connection-pool)"
                    )
                self._counter += 1
            return self

    async def __aexit__(self, *args: Any) -> None:
        self._counter -= 1
        if self._counter == 0:
            assert self._task_group
            tg, self._task_group = self._task_group, None
            tg.cancel_scope.cancel()
            self._reset()
            if self._anchor_reset_token:
                self._anchor_active.reset(self._anchor_reset_token)
            await tg.__aexit__(*args)

    @property
    def task_group(self) -> TaskGroup:
        if self._task_group is None:
            raise RuntimeError(
                "Connection pool is not initialized or has exited. "
                "Make sure you are accessing it after entering the pool's async context manager. "
                "(For more details see https://coredis.readthedocs.org/handbook/connections.html#connection pools)"
            )
        return self._task_group

    @abstractmethod
    async def _initialize(self) -> None:
        """
        To be implemented by the concrete connection pool
        to perform any actions that should only be done on the first
        initialization
        """
        ...

    @abstractmethod
    def _reset(self) -> None:
        """
        To be implemented by the concrete connection pool
        to perform any finalization before the pool context is finally exited
        """
        ...

    @abstractmethod
    async def get_connection(self, **_: Any) -> ConnectionT:
        """
        Gets or create a connection from the pool.
        """
        ...

    @abstractmethod
    def acquire(self, **_: Any) -> AsyncContextManager[ConnectionT]:
        """
        Yields a connection from the pool and releases it back.
        """
        ...

    @abstractmethod
    def release(self, connection: ConnectionT) -> None:
        """
        Release a connection back  to the pool.
        """
        ...

    @abstractmethod
    def disconnect(self) -> None:
        """
        Disconnect all active connections in the pool
        """
        ...

    @classmethod
    def _parse_url(
        cls: type[BaseConnectionPool[ConnectionT]],
        url: str,
        decode_components: bool,
        kwargs: BaseConnectionPoolParamsT,
        pool_param_class: type[BaseConnectionPoolParamsT],
    ) -> tuple[Location | None, BaseConnectionPoolParamsT]:
        parsed_url = urlparse(url)
        query_args = parse_qs(parsed_url.query)
        url_options = pool_param_class(
            **{
                name: cls.URL_QUERY_ARGUMENT_PARSERS.get(name, lambda value: value)(value[0])  # type: ignore
                for name, value in query_args.items()
                if name in pool_param_class.__annotations__ and value
            }
        )
        username: str | None = parsed_url.username
        password: str | None = parsed_url.password
        path: str = parsed_url.path
        hostname: str | None = parsed_url.hostname

        if decode_components:
            username = unquote(username) if username else None
            password = unquote(password) if password else None
            path = unquote(path)
            hostname = unquote(hostname) if hostname else None

        if username:
            url_options["username"] = username
        if password:
            url_options["password"] = password

        location: Location | None = None

        if parsed_url.scheme == "unix":
            location = UnixDomainSocketLocation(path=path)
        else:
            if hostname is not None:
                location = TCPLocation(host=hostname, port=int(parsed_url.port or 6379))
            # If there's a path argument, use it as the db argument if a
            # querystring value wasn't specified

            if "db" not in url_options and path:
                try:
                    url_options["db"] = int(path.replace("/", ""))
                except (AttributeError, ValueError):
                    pass
            if parsed_url.scheme == "rediss" and "ssl_context" not in kwargs:
                ssl_args: dict[str, str | None] = {
                    k: value[0]
                    for k, value in query_args.items()
                    if k
                    in {
                        "ssl_keyfile",
                        "ssl_certfile",
                        "ssl_check_hostname",
                        "ssl_ca_certs",
                        "ssl_cert_reqs",
                    }
                }
                keyfile = ssl_args.get("ssl_keyfile", None)
                certfile = ssl_args.get("ssl_certfile", None)
                check_hostname = query_param_to_bool(ssl_args.get("ssl_check_hostname", None))
                cert_reqs = ssl_args.get("ssl_cert_reqs", None)
                ca_certs = ssl_args.get("ssl_ca_certs", None)
                url_options["ssl_context"] = RedisSSLContext(
                    keyfile, certfile, cert_reqs, ca_certs, check_hostname
                ).get()
            if not url_options.get("db") and "db" in kwargs:
                url_options["db"] = kwargs["db"]
        kwargs.update(url_options)
        return location, kwargs



from __future__ import annotations

from typing import Any, cast

from coredis._utils import nativestr
from coredis.response._callbacks import ResponseCallback
from coredis.response.types import Command
from coredis.typing import (
    AnyStr,
    Hashable,
    ResponsePrimitive,
    ResponseType,
    StringT,
)


class CommandCallback(
    ResponseCallback[
        list[list[ResponsePrimitive | dict[Hashable, ResponseType] | set[Hashable]]],
        dict[str, Command],
    ]
):
    def transform(
        self,
        response: list[list[ResponsePrimitive | dict[Hashable, ResponseType] | set[Hashable]]],
    ) -> dict[str, Command]:
        commands: dict[str, Command] = {}

        for command in response:
            if command:
                detail = cast(list[Any], command)
                name = nativestr(detail[0])

                if len(command) >= 6:
                    commands[name] = {
                        "name": detail[0],
                        "arity": detail[1],
                        "flags": detail[2],
                        "first-key": detail[3],
                        "last-key": detail[4],
                        "step": detail[5],
                        "acl-categories": None,
                        "tips": None,
                        "key-specifications": None,
                        "sub-commands": None,
                    }

                if len(command) >= 7:
                    commands[name]["acl-categories"] = detail[6]

                if len(command) >= 8:
                    commands[name]["tips"] = detail[7]
                    commands[name]["key-specifications"] = detail[8]
                    commands[name]["sub-commands"] = detail[9]

        return commands


class CommandKeyFlagCallback(
    ResponseCallback[list[list[StringT | set[StringT]]], dict[AnyStr, set[AnyStr]]]
):
    def transform(
        self,
        response: list[list[StringT | set[StringT]]],
    ) -> dict[AnyStr, set[AnyStr]]:
        return {cast(AnyStr, k[0]): set(cast(set[AnyStr], k[1])) for k in response}



from __future__ import annotations

import random
from typing import TYPE_CHECKING, Any

from coredis._utils import nativestr
from coredis.connection import BaseConnectionParams, TCPLocation
from coredis.connection._sentinel import SentinelManagedConnection
from coredis.exceptions import PrimaryNotFoundError, ReplicaNotFoundError
from coredis.patterns.cache import AbstractCache
from coredis.typing import StringT, Unpack

from ._basic import ConnectionPool

if TYPE_CHECKING:
    from coredis.client.sentinel import Sentinel


class SentinelConnectionPool(ConnectionPool[SentinelManagedConnection]):
    """
    Sentinel backed connection pool.
    """

    def __init__(
        self,
        service_name: StringT,
        sentinel_manager: Sentinel[Any],
        is_primary: bool = True,
        check_connection: bool = False,
        _cache: AbstractCache | None = None,
        **kwargs: Unpack[BaseConnectionParams],
    ):
        super().__init__(connection_class=SentinelManagedConnection, _cache=_cache, **kwargs)
        self.is_primary = is_primary
        self.service_name = nativestr(service_name)
        self.sentinel_manager = sentinel_manager
        self.check_connection = check_connection
        self.location: TCPLocation | None = None
        self.replicas: list[TCPLocation] = []
        self.replica_counter: int | None = None

    async def _construct_connection(self) -> SentinelManagedConnection:
        if self.is_primary:
            location = await self.get_primary_location()
        else:
            location = await self.get_replica()

        return SentinelManagedConnection(
            location, self.service_name, self.is_primary, **self.connection_kwargs
        )

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}"
            f"<service={self.service_name}"
            f"({'primary' if self.is_primary else 'replica'})>"
        )

    async def get_primary_location(self) -> TCPLocation:
        """
        Returns the location of the primary instance
        """
        primary_location = await self.sentinel_manager.discover_primary(self.service_name)
        location = TCPLocation(primary_location[0], primary_location[1])
        if self.is_primary:
            if self.location != location and self.location is not None:
                # Primary location changed, disconnect all connections in this pool
                self.disconnect()
            self.location = location

        return location

    async def get_replica(self) -> TCPLocation:
        """
        Returns the location of a replica using a round robin approach
        """

        if not self.replicas:
            self.replicas = [
                TCPLocation(*replica)
                for replica in await self.sentinel_manager.discover_replicas(self.service_name)
            ]

        if self.replicas:
            if self.replica_counter is None:
                self.replica_counter = random.randint(0, len(self.replicas) - 1)
            for _ in range(len(self.replicas)):
                self.replica_counter = (self.replica_counter + 1) % len(self.replicas)
                if await self.replicas[self.replica_counter].check():
                    return self.replicas[self.replica_counter]
            # None of the replicas were reachable so we'll need to contact the sentinels
            # again.
            self.replicas.clear()
            self.replica_counter = None

        # If no replicas could be found/reached, lets try a primary for now.
        try:
            return await self.get_primary_location()
        except PrimaryNotFoundError:
            pass
        raise ReplicaNotFoundError(f"No replica found for {self.service_name!r}")



from __future__ import annotations

from collections.abc import Hashable

from .typing import StringT, MutableSet, RedisError

#: Type alias for valid python types that can be represented as json
type JsonType = str | int | float | bool | dict[str, JsonType] | list[JsonType] | None

#: Primitives returned by redis
type ResponsePrimitive = StringT | int | float | bool | None

#: Represents the total structure of any response for any redis command.
type ResponseType = (
    ResponsePrimitive
    | RedisError
    | list[ResponseType]
    | MutableSet[Hashable]
    | dict[Hashable, ResponseType]
)



from __future__ import annotations

from collections import deque
from typing import Any, overload

from anyio import Event, Lock, create_task_group

from coredis.typing import Awaitable, Generic, TypeVar

T1 = TypeVar("T1")
T2 = TypeVar("T2")
T3 = TypeVar("T3")
T4 = TypeVar("T4")
T5 = TypeVar("T5")
T6 = TypeVar("T6")


@overload
async def gather(
    awaitable1: Awaitable[T1],
    awaitable2: Awaitable[T2],
    /,
    *,
    return_exceptions: bool = False,
) -> tuple[T1, T2]: ...


@overload
async def gather(
    awaitable1: Awaitable[T1],
    awaitable2: Awaitable[T2],
    awaitable3: Awaitable[T3],
    /,
    *,
    return_exceptions: bool = False,
) -> tuple[T1, T2, T3]: ...


@overload
async def gather(
    awaitable1: Awaitable[T1],
    awaitable2: Awaitable[T2],
    awaitable3: Awaitable[T3],
    awaitable4: Awaitable[T4],
    /,
    *,
    return_exceptions: bool = False,
) -> tuple[T1, T2, T3, T4]: ...


@overload
async def gather(
    awaitable1: Awaitable[T1],
    awaitable2: Awaitable[T2],
    awaitable3: Awaitable[T3],
    awaitable4: Awaitable[T4],
    awaitable5: Awaitable[T5],
    /,
    *,
    return_exceptions: bool = False,
) -> tuple[T1, T2, T3, T4, T5]: ...


@overload
async def gather(
    awaitable1: Awaitable[T1],
    awaitable2: Awaitable[T2],
    awaitable3: Awaitable[T3],
    awaitable4: Awaitable[T4],
    awaitable5: Awaitable[T5],
    awaitable6: Awaitable[T6],
    /,
    *,
    return_exceptions: bool = False,
) -> tuple[T1, T2, T3, T4, T5, T6]: ...


@overload
async def gather(
    *awaitables: Awaitable[T1],
    return_exceptions: bool = False,
) -> tuple[T1, ...]: ...


async def gather(*awaitables: Awaitable[Any], return_exceptions: bool = False) -> tuple[Any, ...]:
    if not awaitables:
        return ()
    results: list[Any] = [None] * len(awaitables)

    async def runner(awaitable: Awaitable[Any], i: int) -> None:
        try:
            results[i] = await awaitable
        except Exception as exc:
            if not return_exceptions:
                raise
            results[i] = exc

    async with create_task_group() as tg:
        for i, awaitable in enumerate(awaitables):
            tg.start_soon(runner, awaitable, i)

    return tuple(results)


class QueueEmpty(Exception): ...


class QueueFull(Exception): ...


class Queue(Generic[T1]):
    """
    Generic LIFO queue (stack) for use with connections.
    """

    def __init__(self, maxsize: int = 0):
        self._maxsize = maxsize
        self._queue: deque[T1 | None] = deque(
            [None for _ in range(self._maxsize)], maxlen=self._maxsize
        )
        self._getters: deque[Event] = deque()
        self._putters: deque[Event] = deque()
        self._lock = Lock()

    def empty(self) -> bool:
        return not self._queue

    def full(self) -> bool:
        return self._maxsize > 0 and len(self._queue) >= self._maxsize

    async def put(self, item: T1 | None) -> None:
        async with self._lock:
            while self.full():
                ev = Event()
                self._putters.append(ev)
                await ev.wait()
            self._queue.append(item)
            if self._getters:
                self._getters.popleft().set()

    def put_nowait(self, item: T1 | None) -> None:
        if self.full():
            raise QueueFull()
        self._queue.append(item)
        if self._getters:
            ev = self._getters.popleft()
            ev.set()

    def append_nowait(self, item: T1 | None) -> None:
        if self.full():
            raise QueueFull()
        self._queue.appendleft(item)
        if self._getters:
            ev = self._getters.popleft()
            ev.set()

    async def get(self) -> T1 | None:
        async with self._lock:
            while self.empty():
                ev = Event()
                self._getters.append(ev)
                await ev.wait()
            item = self._queue.pop()
            if self._putters and not self.full():
                self._putters.popleft().set()

            return item

    def get_nowait(self) -> T1 | None:
        if self.empty():
            raise QueueEmpty()
        item = self._queue.pop()
        if self._putters and not self.full():
            self._putters.popleft().set()

        return item

    def remove(self, item: T1) -> None:
        self._queue.remove(item)

    def __len__(self) -> int:
        return len(self._queue)

    def __contains__(self, item: T1) -> bool:
        return item in self._queue



from __future__ import annotations

from coredis._utils import b, hash_slot
from coredis.typing import Callable, ClassVar, RedisValueT


class KeySpec:
    READONLY_VARIANTS: ClassVar[
        dict[
            bytes,
            Callable[
                [tuple[RedisValueT, ...]],
                list[tuple[tuple[RedisValueT, ...], int | None, int | None]],
            ],
        ]
    ] = {
        b"BITCOUNT": lambda args: [((args[1],), 1, 1)],
        b"BITFIELD_RO": lambda args: [((args[1],), 1, 1)],
        b"BITOP": lambda args: [(args[3 : len(args)], 3, len(args) - 1)],
        b"BITPOS": lambda args: [((args[1],), 1, 1)],
        b"COPY": lambda args: [((args[1],), 1, 1)],
        b"DIGEST": lambda args: [((args[1],), 1, 1)],
        b"DUMP": lambda args: [((args[1],), 1, 1)],
        b"EVALSHA_RO": lambda args: [
            (args[3 : 3 + (int(args[2]) * 1)], 3, 3 + (int(args[2]) * 1) - 1)
        ],
        b"EVAL_RO": lambda args: [
            (args[3 : 3 + (int(args[2]) * 1)], 3, 3 + (int(args[2]) * 1) - 1)
        ],
        b"EXISTS": lambda args: [(args[1 : len(args)], 1, len(args) - 1)],
        b"EXPIRETIME": lambda args: [((args[1],), 1, 1)],
        b"FCALL_RO": lambda args: [
            (args[3 : 3 + (int(args[2]) * 1)], 3, 3 + (int(args[2]) * 1) - 1)
        ],
        b"GEODIST": lambda args: [((args[1],), 1, 1)],
        b"GEOHASH": lambda args: [((args[1],), 1, 1)],
        b"GEOPOS": lambda args: [((args[1],), 1, 1)],
        b"GEORADIUS": lambda args: [((args[1],), 1, 1)],
        b"GEORADIUSBYMEMBER": lambda args: [((args[1],), 1, 1)],
        b"GEORADIUSBYMEMBER_RO": lambda args: [((args[1],), 1, 1)],
        b"GEORADIUS_RO": lambda args: [((args[1],), 1, 1)],
        b"GEOSEARCH": lambda args: [((args[1],), 1, 1)],
        b"GEOSEARCHSTORE": lambda args: [((args[2],), 2, 2)],
        b"GET": lambda args: [((args[1],), 1, 1)],
        b"GETBIT": lambda args: [((args[1],), 1, 1)],
        b"GETRANGE": lambda args: [((args[1],), 1, 1)],
        b"HEXISTS": lambda args: [((args[1],), 1, 1)],
        b"HEXPIRETIME": lambda args: [((args[1],), 1, 1)],
        b"HGET": lambda args: [((args[1],), 1, 1)],
        b"HGETALL": lambda args: [((args[1],), 1, 1)],
        b"HKEYS": lambda args: [((args[1],), 1, 1)],
        b"HLEN": lambda args: [((args[1],), 1, 1)],
        b"HMGET": lambda args: [((args[1],), 1, 1)],
        b"HPEXPIRETIME": lambda args: [((args[1],), 1, 1)],
        b"HPTTL": lambda args: [((args[1],), 1, 1)],
        b"HRANDFIELD": lambda args: [((args[1],), 1, 1)],
        b"HSCAN": lambda args: [((args[1],), 1, 1)],
        b"HSTRLEN": lambda args: [((args[1],), 1, 1)],
        b"HTTL": lambda args: [((args[1],), 1, 1)],
        b"HVALS": lambda args: [((args[1],), 1, 1)],
        b"LCS": lambda args: [(args[1:3], 1, 3 - 1)],
        b"LINDEX": lambda args: [((args[1],), 1, 1)],
        b"LLEN": lambda args: [((args[1],), 1, 1)],
        b"LPOS": lambda args: [((args[1],), 1, 1)],
        b"LRANGE": lambda args: [((args[1],), 1, 1)],
        b"MEMORY USAGE": lambda args: [((args[1],), 1, 1)],
        b"MGET": lambda args: [(args[1 : len(args)], 1, len(args) - 1)],
        b"OBJECT ENCODING": lambda args: [((args[1],), 1, 1)],
        b"OBJECT FREQ": lambda args: [((args[1],), 1, 1)],
        b"OBJECT IDLETIME": lambda args: [((args[1],), 1, 1)],
        b"OBJECT REFCOUNT": lambda args: [((args[1],), 1, 1)],
        b"PEXPIRETIME": lambda args: [((args[1],), 1, 1)],
        b"PFMERGE": lambda args: [(args[2 : len(args)], 2, len(args) - 1)],
        b"PTTL": lambda args: [((args[1],), 1, 1)],
        b"SCARD": lambda args: [((args[1],), 1, 1)],
        b"SDIFF": lambda args: [(args[1 : len(args)], 1, len(args) - 1)],
        b"SDIFFSTORE": lambda args: [(args[2 : len(args)], 2, len(args) - 1)],
        b"SINTER": lambda args: [(args[1 : len(args)], 1, len(args) - 1)],
        b"SINTERCARD": lambda args: [
            (args[2 : 2 + (int(args[1]) * 1)], 2, 2 + (int(args[1]) * 1) - 1)
        ],
        b"SINTERSTORE": lambda args: [(args[2 : len(args)], 2, len(args) - 1)],
        b"SISMEMBER": lambda args: [((args[1],), 1, 1)],
        b"SMEMBERS": lambda args: [((args[1],), 1, 1)],
        b"SMISMEMBER": lambda args: [((args[1],), 1, 1)],
        b"SORT": lambda args: [((args[1],), 1, 1)],
        b"SORT_RO": lambda args: [((args[1],), 1, 1)],
        b"SRANDMEMBER": lambda args: [((args[1],), 1, 1)],
        b"SSCAN": lambda args: [((args[1],), 1, 1)],
        b"STRLEN": lambda args: [((args[1],), 1, 1)],
        b"SUBSTR": lambda args: [((args[1],), 1, 1)],
        b"SUNION": lambda args: [(args[1 : len(args)], 1, len(args) - 1)],
        b"SUNIONSTORE": lambda args: [(args[2 : len(args)], 2, len(args) - 1)],
        b"TOUCH": lambda args: [(args[1 : len(args)], 1, len(args) - 1)],
        b"TTL": lambda args: [((args[1],), 1, 1)],
        b"TYPE": lambda args: [((args[1],), 1, 1)],
        b"WATCH": lambda args: [(args[1 : len(args)], 1, len(args) - 1)],
        b"XINFO CONSUMERS": lambda args: [((args[1],), 1, 1)],
        b"XINFO GROUPS": lambda args: [((args[1],), 1, 1)],
        b"XINFO STREAM": lambda args: [((args[1],), 1, 1)],
        b"XLEN": lambda args: [((args[1],), 1, 1)],
        b"XPENDING": lambda args: [((args[1],), 1, 1)],
        b"XRANGE": lambda args: [((args[1],), 1, 1)],
        b"XREAD": lambda args: [
            (
                (
                    lambda kwpos: (
                        tuple(args[1 + kwpos : len(args) - (len(args) - (kwpos + 1)) // 2]),
                        1 + kwpos,
                        len(args) - (len(args) - (kwpos + 1)) // 2 - 1,
                    )
                )(args.index(b"STREAMS", 1))
                if b"STREAMS" in args
                else ((), None, None)
            )
        ],
        b"XREADGROUP": lambda args: [
            (
                (
                    lambda kwpos: (
                        tuple(args[1 + kwpos : len(args) - (len(args) - (kwpos + 1)) // 2]),
                        1 + kwpos,
                        len(args) - (len(args) - (kwpos + 1)) // 2 - 1,
                    )
                )(args.index(b"STREAMS", 4))
                if b"STREAMS" in args
                else ((), None, None)
            )
        ],
        b"XREVRANGE": lambda args: [((args[1],), 1, 1)],
        b"ZCARD": lambda args: [((args[1],), 1, 1)],
        b"ZCOUNT": lambda args: [((args[1],), 1, 1)],
        b"ZDIFF": lambda args: [(args[2 : 2 + (int(args[1]) * 1)], 2, 2 + (int(args[1]) * 1) - 1)],
        b"ZDIFFSTORE": lambda args: [
            (args[3 : 3 + (int(args[2]) * 1)], 3, 3 + (int(args[2]) * 1) - 1)
        ],
        b"ZINTER": lambda args: [(args[2 : 2 + (int(args[1]) * 1)], 2, 2 + (int(args[1]) * 1) - 1)],
        b"ZINTERCARD": lambda args: [
            (args[2 : 2 + (int(args[1]) * 1)], 2, 2 + (int(args[1]) * 1) - 1)
        ],
        b"ZINTERSTORE": lambda args: [
            (args[3 : 3 + (int(args[2]) * 1)], 3, 3 + (int(args[2]) * 1) - 1)
        ],
        b"ZLEXCOUNT": lambda args: [((args[1],), 1, 1)],
        b"ZMSCORE": lambda args: [((args[1],), 1, 1)],
        b"ZRANDMEMBER": lambda args: [((args[1],), 1, 1)],
        b"ZRANGE": lambda args: [((args[1],), 1, 1)],
        b"ZRANGEBYLEX": lambda args: [((args[1],), 1, 1)],
        b"ZRANGEBYSCORE": lambda args: [((args[1],), 1, 1)],
        b"ZRANGESTORE": lambda args: [((args[2],), 2, 2)],
        b"ZRANK": lambda args: [((args[1],), 1, 1)],
        b"ZREVRANGE": lambda args: [((args[1],), 1, 1)],
        b"ZREVRANGEBYLEX": lambda args: [((args[1],), 1, 1)],
        b"ZREVRANGEBYSCORE": lambda args: [((args[1],), 1, 1)],
        b"ZREVRANK": lambda args: [((args[1],), 1, 1)],
        b"ZSCAN": lambda args: [((args[1],), 1, 1)],
        b"ZSCORE": lambda args: [((args[1],), 1, 1)],
        b"ZUNION": lambda args: [(args[2 : 2 + (int(args[1]) * 1)], 2, 2 + (int(args[1]) * 1) - 1)],
        b"ZUNIONSTORE": lambda args: [
            (args[3 : 3 + (int(args[2]) * 1)], 3, 3 + (int(args[2]) * 1) - 1)
        ],
    }
    COMMANDS: ClassVar[
        dict[
            bytes,
            Callable[
                [tuple[RedisValueT, ...]],
                list[tuple[tuple[RedisValueT, ...], int | None, int | None]],
            ],
        ]
    ] = {
        b"VADD": lambda args: [((args[1],), 1, 1)],
        b"VREM": lambda args: [((args[1],), 1, 1)],
        b"VSIM": lambda args: [((args[1],), 1, 1)],
        b"VDIM": lambda args: [((args[1],), 1, 1)],
        b"VCARD": lambda args: [((args[1],), 1, 1)],
        b"VEMB": lambda args: [((args[1],), 1, 1)],
        b"VLINKS": lambda args: [((args[1],), 1, 1)],
        b"VINFO": lambda args: [((args[1],), 1, 1)],
        b"VSETATTR": lambda args: [((args[1],), 1, 1)],
        b"VGETATTR": lambda args: [((args[1],), 1, 1)],
        b"VRANDMEMBER": lambda args: [((args[1],), 1, 1)],
        b"VISMEMBER": lambda args: [((args[1],), 1, 1)],
        b"VRANGE": lambda args: [((args[1],), 1, 1)],
        b"OBJECT": lambda args: [((args[2],), 2, 2)],
        b"DEBUG OBJECT": lambda args: [((args[1],), 1, 1)],
        b"APPEND": lambda args: [((args[1],), 1, 1)],
        b"BITCOUNT": lambda args: [((args[1],), 1, 1)],
        b"BITFIELD": lambda args: [((args[1],), 1, 1)],
        b"BITFIELD_RO": lambda args: [((args[1],), 1, 1)],
        b"BITOP": lambda args: [((args[2],), 2, 2), (args[3 : len(args)], 3, len(args) - 1)],
        b"BITPOS": lambda args: [((args[1],), 1, 1)],
        b"BLMOVE": lambda args: [((args[1],), 1, 1), ((args[2],), 2, 2)],
        b"BLMPOP": lambda args: [(args[3 : 3 + (int(args[2]) * 1)], 3, 3 + (int(args[2]) * 1) - 1)],
        b"BLPOP": lambda args: [(args[1 : len(args) - 1], 1, len(args) - 1 - 1)],
        b"BRPOP": lambda args: [(args[1 : len(args) - 1], 1, len(args) - 1 - 1)],
        b"BRPOPLPUSH": lambda args: [((args[1],), 1, 1), ((args[2],), 2, 2)],
        b"BZMPOP": lambda args: [(args[3 : 3 + (int(args[2]) * 1)], 3, 3 + (int(args[2]) * 1) - 1)],
        b"BZPOPMAX": lambda args: [(args[1 : len(args) - 1], 1, len(args) - 1 - 1)],
        b"BZPOPMIN": lambda args: [(args[1 : len(args) - 1], 1, len(args) - 1 - 1)],
        b"COPY": lambda args: [((args[1],), 1, 1), ((args[2],), 2, 2)],
        b"DECR": lambda args: [((args[1],), 1, 1)],
        b"DECRBY": lambda args: [((args[1],), 1, 1)],
        b"DEL": lambda args: [(args[1 : len(args)], 1, len(args) - 1)],
        b"DELEX": lambda args: [((args[1],), 1, 1)],
        b"DIGEST": lambda args: [((args[1],), 1, 1)],
        b"DUMP": lambda args: [((args[1],), 1, 1)],
        b"EVAL": lambda args: [(args[3 : 3 + (int(args[2]) * 1)], 3, 3 + (int(args[2]) * 1) - 1)],
        b"EVALSHA": lambda args: [
            (args[3 : 3 + (int(args[2]) * 1)], 3, 3 + (int(args[2]) * 1) - 1)
        ],
        b"EVALSHA_RO": lambda args: [
            (args[3 : 3 + (int(args[2]) * 1)], 3, 3 + (int(args[2]) * 1) - 1)
        ],
        b"EVAL_RO": lambda args: [
            (args[3 : 3 + (int(args[2]) * 1)], 3, 3 + (int(args[2]) * 1) - 1)
        ],
        b"EXISTS": lambda args: [(args[1 : len(args)], 1, len(args) - 1)],
        b"EXPIRE": lambda args: [((args[1],), 1, 1)],
        b"EXPIREAT": lambda args: [((args[1],), 1, 1)],
        b"EXPIRETIME": lambda args: [((args[1],), 1, 1)],
        b"FCALL": lambda args: [(args[3 : 3 + (int(args[2]) * 1)], 3, 3 + (int(args[2]) * 1) - 1)],
        b"FCALL_RO": lambda args: [
            (args[3 : 3 + (int(args[2]) * 1)], 3, 3 + (int(args[2]) * 1) - 1)
        ],
        b"GEOADD": lambda args: [((args[1],), 1, 1)],
        b"GEODIST": lambda args: [((args[1],), 1, 1)],
        b"GEOHASH": lambda args: [((args[1],), 1, 1)],
        b"GEOPOS": lambda args: [((args[1],), 1, 1)],
        b"GEORADIUS": lambda args: [
            ((args[1],), 1, 1),
            (
                (
                    lambda kwpos: (
                        tuple(args[1 + kwpos : kwpos + 1 + 1]),
                        1 + kwpos,
                        kwpos + 1 + 1 - 1,
                    )
                )(args.index(b"STORE", 6))
                if b"STORE" in args
                else ((), None, None)
            ),
            (
                (
                    lambda kwpos: (
                        tuple(args[1 + kwpos : kwpos + 1 + 1]),
                        1 + kwpos,
                        kwpos + 1 + 1 - 1,
                    )
                )(args.index(b"STOREDIST", 6))
                if b"STOREDIST" in args
                else ((), None, None)
            ),
        ],
        b"GEORADIUSBYMEMBER": lambda args: [
            ((args[1],), 1, 1),
            (
                (
                    lambda kwpos: (
                        tuple(args[1 + kwpos : kwpos + 1 + 1]),
                        1 + kwpos,
                        kwpos + 1 + 1 - 1,
                    )
                )(args.index(b"STORE", 5))
                if b"STORE" in args
                else ((), None, None)
            ),
            (
                (
                    lambda kwpos: (
                        tuple(args[1 + kwpos : kwpos + 1 + 1]),
                        1 + kwpos,
                        kwpos + 1 + 1 - 1,
                    )
                )(args.index(b"STOREDIST", 5))
                if b"STOREDIST" in args
                else ((), None, None)
            ),
        ],
        b"GEORADIUSBYMEMBER_RO": lambda args: [((args[1],), 1, 1)],
        b"GEORADIUS_RO": lambda args: [((args[1],), 1, 1)],
        b"GEOSEARCH": lambda args: [((args[1],), 1, 1)],
        b"GEOSEARCHSTORE": lambda args: [((args[1],), 1, 1), ((args[2],), 2, 2)],
        b"GET": lambda args: [((args[1],), 1, 1)],
        b"GETBIT": lambda args: [((args[1],), 1, 1)],
        b"GETDEL": lambda args: [((args[1],), 1, 1)],
        b"GETEX": lambda args: [((args[1],), 1, 1)],
        b"GETRANGE": lambda args: [((args[1],), 1, 1)],
        b"GETSET": lambda args: [((args[1],), 1, 1)],
        b"HDEL": lambda args: [((args[1],), 1, 1)],
        b"HEXISTS": lambda args: [((args[1],), 1, 1)],
        b"HEXPIRE": lambda args: [((args[1],), 1, 1)],
        b"HEXPIREAT": lambda args: [((args[1],), 1, 1)],
        b"HEXPIRETIME": lambda args: [((args[1],), 1, 1)],
        b"HGET": lambda args: [((args[1],), 1, 1)],
        b"HGETALL": lambda args: [((args[1],), 1, 1)],
        b"HGETDEL": lambda args: [((args[1],), 1, 1)],
        b"HGETEX": lambda args: [((args[1],), 1, 1)],
        b"HINCRBY": lambda args: [((args[1],), 1, 1)],
        b"HINCRBYFLOAT": lambda args: [((args[1],), 1, 1)],
        b"HKEYS": lambda args: [((args[1],), 1, 1)],
        b"HLEN": lambda args: [((args[1],), 1, 1)],
        b"HMGET": lambda args: [((args[1],), 1, 1)],
        b"HMSET": lambda args: [((args[1],), 1, 1)],
        b"HPERSIST": lambda args: [((args[1],), 1, 1)],
        b"HPEXPIRE": lambda args: [((args[1],), 1, 1)],
        b"HPEXPIREAT": lambda args: [((args[1],), 1, 1)],
        b"HPEXPIRETIME": lambda args: [((args[1],), 1, 1)],
        b"HPTTL": lambda args: [((args[1],), 1, 1)],
        b"HRANDFIELD": lambda args: [((args[1],), 1, 1)],
        b"HSCAN": lambda args: [((args[1],), 1, 1)],
        b"HSET": lambda args: [((args[1],), 1, 1)],
        b"HSETEX": lambda args: [((args[1],), 1, 1)],
        b"HSETNX": lambda args: [((args[1],), 1, 1)],
        b"HSTRLEN": lambda args: [((args[1],), 1, 1)],
        b"HTTL": lambda args: [((args[1],), 1, 1)],
        b"HVALS": lambda args: [((args[1],), 1, 1)],
        b"INCR": lambda args: [((args[1],), 1, 1)],
        b"INCRBY": lambda args: [((args[1],), 1, 1)],
        b"INCRBYFLOAT": lambda args: [((args[1],), 1, 1)],
        b"LCS": lambda args: [(args[1:3], 1, 3 - 1)],
        b"LINDEX": lambda args: [((args[1],), 1, 1)],
        b"LINSERT": lambda args: [((args[1],), 1, 1)],
        b"LLEN": lambda args: [((args[1],), 1, 1)],
        b"LMOVE": lambda args: [((args[1],), 1, 1), ((args[2],), 2, 2)],
        b"LMPOP": lambda args: [(args[2 : 2 + (int(args[1]) * 1)], 2, 2 + (int(args[1]) * 1) - 1)],
        b"LPOP": lambda args: [((args[1],), 1, 1)],
        b"LPOS": lambda args: [((args[1],), 1, 1)],
        b"LPUSH": lambda args: [((args[1],), 1, 1)],
        b"LPUSHX": lambda args: [((args[1],), 1, 1)],
        b"LRANGE": lambda args: [((args[1],), 1, 1)],
        b"LREM": lambda args: [((args[1],), 1, 1)],
        b"LSET": lambda args: [((args[1],), 1, 1)],
        b"LTRIM": lambda args: [((args[1],), 1, 1)],
        b"MEMORY USAGE": lambda args: [((args[1],), 1, 1)],
        b"MGET": lambda args: [(args[1 : len(args)], 1, len(args) - 1)],
        b"MIGRATE": lambda args: [
            ((args[3],), 3, 3),
            (
                (lambda kwpos: (tuple(args[1 + kwpos : len(args)]), 1 + kwpos, len(args) - 1))(
                    len(args) - list(reversed(args)).index(b"KEYS", 1) - 1
                )
                if b"KEYS" in args
                else ((), None, None)
            ),
        ],
        b"MOVE": lambda args: [((args[1],), 1, 1)],
        b"MSET": lambda args: [(args[1 : len(args) : 2], 1, len(args) - 2)],
        b"MSETEX": lambda args: [
            (args[2 : 2 + (int(args[1]) * 2) : 2], 2, 2 + (int(args[1]) * 2) - 2)
        ],
        b"MSETNX": lambda args: [(args[1 : len(args) : 2], 1, len(args) - 2)],
        b"OBJECT ENCODING": lambda args: [((args[1],), 1, 1)],
        b"OBJECT FREQ": lambda args: [((args[1],), 1, 1)],
        b"OBJECT IDLETIME": lambda args: [((args[1],), 1, 1)],
        b"OBJECT REFCOUNT": lambda args: [((args[1],), 1, 1)],
        b"PERSIST": lambda args: [((args[1],), 1, 1)],
        b"PEXPIRE": lambda args: [((args[1],), 1, 1)],
        b"PEXPIREAT": lambda args: [((args[1],), 1, 1)],
        b"PEXPIRETIME": lambda args: [((args[1],), 1, 1)],
        b"PFADD": lambda args: [((args[1],), 1, 1)],
        b"PFCOUNT": lambda args: [(args[1 : len(args)], 1, len(args) - 1)],
        b"PFDEBUG": lambda args: [((args[2],), 2, 2)],
        b"PFMERGE": lambda args: [((args[1],), 1, 1), (args[2 : len(args)], 2, len(args) - 1)],
        b"PSETEX": lambda args: [((args[1],), 1, 1)],
        b"PTTL": lambda args: [((args[1],), 1, 1)],
        b"RENAME": lambda args: [((args[1],), 1, 1), ((args[2],), 2, 2)],
        b"RENAMENX": lambda args: [((args[1],), 1, 1), ((args[2],), 2, 2)],
        b"RESTORE": lambda args: [((args[1],), 1, 1)],
        b"RESTORE-ASKING": lambda args: [((args[1],), 1, 1)],
        b"RPOP": lambda args: [((args[1],), 1, 1)],
        b"RPOPLPUSH": lambda args: [((args[1],), 1, 1), ((args[2],), 2, 2)],
        b"RPUSH": lambda args: [((args[1],), 1, 1)],
        b"RPUSHX": lambda args: [((args[1],), 1, 1)],
        b"SADD": lambda args: [((args[1],), 1, 1)],
        b"SCARD": lambda args: [((args[1],), 1, 1)],
        b"SDIFF": lambda args: [(args[1 : len(args)], 1, len(args) - 1)],
        b"SDIFFSTORE": lambda args: [((args[1],), 1, 1), (args[2 : len(args)], 2, len(args) - 1)],
        b"SET": lambda args: [((args[1],), 1, 1)],
        b"SETBIT": lambda args: [((args[1],), 1, 1)],
        b"SETEX": lambda args: [((args[1],), 1, 1)],
        b"SETNX": lambda args: [((args[1],), 1, 1)],
        b"SETRANGE": lambda args: [((args[1],), 1, 1)],
        b"SINTER": lambda args: [(args[1 : len(args)], 1, len(args) - 1)],
        b"SINTERCARD": lambda args: [
            (args[2 : 2 + (int(args[1]) * 1)], 2, 2 + (int(args[1]) * 1) - 1)
        ],
        b"SINTERSTORE": lambda args: [((args[1],), 1, 1), (args[2 : len(args)], 2, len(args) - 1)],
        b"SISMEMBER": lambda args: [((args[1],), 1, 1)],
        b"SMEMBERS": lambda args: [((args[1],), 1, 1)],
        b"SMISMEMBER": lambda args: [((args[1],), 1, 1)],
        b"SMOVE": lambda args: [((args[1],), 1, 1), ((args[2],), 2, 2)],
        b"SORT": lambda args: [((args[1],), 1, 1)],
        b"SORT_RO": lambda args: [((args[1],), 1, 1)],
        b"SPOP": lambda args: [((args[1],), 1, 1)],
        b"SPUBLISH": lambda args: [((args[1],), 1, 1)],
        b"SRANDMEMBER": lambda args: [((args[1],), 1, 1)],
        b"SREM": lambda args: [((args[1],), 1, 1)],
        b"SSCAN": lambda args: [((args[1],), 1, 1)],
        b"SSUBSCRIBE": lambda args: [(args[1 : len(args)], 1, len(args) - 1)],
        b"STRLEN": lambda args: [((args[1],), 1, 1)],
        b"SUBSTR": lambda args: [((args[1],), 1, 1)],
        b"SUNION": lambda args: [(args[1 : len(args)], 1, len(args) - 1)],
        b"SUNIONSTORE": lambda args: [((args[1],), 1, 1), (args[2 : len(args)], 2, len(args) - 1)],
        b"SUNSUBSCRIBE": lambda args: [(args[1 : len(args)], 1, len(args) - 1)],
        b"TOUCH": lambda args: [(args[1 : len(args)], 1, len(args) - 1)],
        b"TTL": lambda args: [((args[1],), 1, 1)],
        b"TYPE": lambda args: [((args[1],), 1, 1)],
        b"UNLINK": lambda args: [(args[1 : len(args)], 1, len(args) - 1)],
        b"WATCH": lambda args: [(args[1 : len(args)], 1, len(args) - 1)],
        b"XACK": lambda args: [((args[1],), 1, 1)],
        b"XACKDEL": lambda args: [((args[1],), 1, 1)],
        b"XADD": lambda args: [((args[1],), 1, 1)],
        b"XAUTOCLAIM": lambda args: [((args[1],), 1, 1)],
        b"XCFGSET": lambda args: [((args[1],), 1, 1)],
        b"XCLAIM": lambda args: [((args[1],), 1, 1)],
        b"XDEL": lambda args: [((args[1],), 1, 1)],
        b"XDELEX": lambda args: [((args[1],), 1, 1)],
        b"XGROUP CREATE": lambda args: [((args[1],), 1, 1)],
        b"XGROUP CREATECONSUMER": lambda args: [((args[1],), 1, 1)],
        b"XGROUP DELCONSUMER": lambda args: [((args[1],), 1, 1)],
        b"XGROUP DESTROY": lambda args: [((args[1],), 1, 1)],
        b"XGROUP SETID": lambda args: [((args[1],), 1, 1)],
        b"XINFO CONSUMERS": lambda args: [((args[1],), 1, 1)],
        b"XINFO GROUPS": lambda args: [((args[1],), 1, 1)],
        b"XINFO STREAM": lambda args: [((args[1],), 1, 1)],
        b"XLEN": lambda args: [((args[1],), 1, 1)],
        b"XPENDING": lambda args: [((args[1],), 1, 1)],
        b"XRANGE": lambda args: [((args[1],), 1, 1)],
        b"XREAD": lambda args: [
            (
                (
                    lambda kwpos: (
                        tuple(args[1 + kwpos : len(args) - (len(args) - (kwpos + 1)) // 2]),
                        1 + kwpos,
                        len(args) - (len(args) - (kwpos + 1)) // 2 - 1,
                    )
                )(args.index(b"STREAMS", 1))
                if b"STREAMS" in args
                else ((), None, None)
            )
        ],
        b"XREADGROUP": lambda args: [
            (
                (
                    lambda kwpos: (
                        tuple(args[1 + kwpos : len(args) - (len(args) - (kwpos + 1)) // 2]),
                        1 + kwpos,
                        len(args) - (len(args) - (kwpos + 1)) // 2 - 1,
                    )
                )(args.index(b"STREAMS", 4))
                if b"STREAMS" in args
                else ((), None, None)
            )
        ],
        b"XREVRANGE": lambda args: [((args[1],), 1, 1)],
        b"XSETID": lambda args: [((args[1],), 1, 1)],
        b"XTRIM": lambda args: [((args[1],), 1, 1)],
        b"ZADD": lambda args: [((args[1],), 1, 1)],
        b"ZCARD": lambda args: [((args[1],), 1, 1)],
        b"ZCOUNT": lambda args: [((args[1],), 1, 1)],
        b"ZDIFF": lambda args: [(args[2 : 2 + (int(args[1]) * 1)], 2, 2 + (int(args[1]) * 1) - 1)],
        b"ZDIFFSTORE": lambda args: [
            ((args[1],), 1, 1),
            (args[3 : 3 + (int(args[2]) * 1)], 3, 3 + (int(args[2]) * 1) - 1),
        ],
        b"ZINCRBY": lambda args: [((args[1],), 1, 1)],
        b"ZINTER": lambda args: [(args[2 : 2 + (int(args[1]) * 1)], 2, 2 + (int(args[1]) * 1) - 1)],
        b"ZINTERCARD": lambda args: [
            (args[2 : 2 + (int(args[1]) * 1)], 2, 2 + (int(args[1]) * 1) - 1)
        ],
        b"ZINTERSTORE": lambda args: [
            ((args[1],), 1, 1),
            (args[3 : 3 + (int(args[2]) * 1)], 3, 3 + (int(args[2]) * 1) - 1),
        ],
        b"ZLEXCOUNT": lambda args: [((args[1],), 1, 1)],
        b"ZMPOP": lambda args: [(args[2 : 2 + (int(args[1]) * 1)], 2, 2 + (int(args[1]) * 1) - 1)],
        b"ZMSCORE": lambda args: [((args[1],), 1, 1)],
        b"ZPOPMAX": lambda args: [((args[1],), 1, 1)],
        b"ZPOPMIN": lambda args: [((args[1],), 1, 1)],
        b"ZRANDMEMBER": lambda args: [((args[1],), 1, 1)],
        b"ZRANGE": lambda args: [((args[1],), 1, 1)],
        b"ZRANGEBYLEX": lambda args: [((args[1],), 1, 1)],
        b"ZRANGEBYSCORE": lambda args: [((args[1],), 1, 1)],
        b"ZRANGESTORE": lambda args: [((args[1],), 1, 1), ((args[2],), 2, 2)],
        b"ZRANK": lambda args: [((args[1],), 1, 1)],
        b"ZREM": lambda args: [((args[1],), 1, 1)],
        b"ZREMRANGEBYLEX": lambda args: [((args[1],), 1, 1)],
        b"ZREMRANGEBYRANK": lambda args: [((args[1],), 1, 1)],
        b"ZREMRANGEBYSCORE": lambda args: [((args[1],), 1, 1)],
        b"ZREVRANGE": lambda args: [((args[1],), 1, 1)],
        b"ZREVRANGEBYLEX": lambda args: [((args[1],), 1, 1)],
        b"ZREVRANGEBYSCORE": lambda args: [((args[1],), 1, 1)],
        b"ZREVRANK": lambda args: [((args[1],), 1, 1)],
        b"ZSCAN": lambda args: [((args[1],), 1, 1)],
        b"ZSCORE": lambda args: [((args[1],), 1, 1)],
        b"ZUNION": lambda args: [(args[2 : 2 + (int(args[1]) * 1)], 2, 2 + (int(args[1]) * 1) - 1)],
        b"ZUNIONSTORE": lambda args: [
            ((args[1],), 1, 1),
            (args[3 : 3 + (int(args[2]) * 1)], 3, 3 + (int(args[2]) * 1) - 1),
        ],
        b"JSON.DEBUG MEMORY": lambda args: [((args[1],), 1, 1)],
        b"JSON.DEL": lambda args: [((args[1],), 1, 1)],
        b"JSON.FORGET": lambda args: [((args[1],), 1, 1)],
        b"JSON.GET": lambda args: [((args[1],), 1, 1)],
        b"JSON.SET": lambda args: [((args[1],), 1, 1)],
        b"JSON.NUMINCRBY": lambda args: [((args[1],), 1, 1)],
        b"JSON.STRAPPEND": lambda args: [((args[1],), 1, 1)],
        b"JSON.STRLEN": lambda args: [((args[1],), 1, 1)],
        b"JSON.ARRAPPEND": lambda args: [((args[1],), 1, 1)],
        b"JSON.ARRINDEX": lambda args: [((args[1],), 1, 1)],
        b"JSON.ARRINSERT": lambda args: [((args[1],), 1, 1)],
        b"JSON.ARRLEN": lambda args: [((args[1],), 1, 1)],
        b"JSON.ARRPOP": lambda args: [((args[1],), 1, 1)],
        b"JSON.ARRTRIM": lambda args: [((args[1],), 1, 1)],
        b"JSON.OBJKEYS": lambda args: [((args[1],), 1, 1)],
        b"JSON.OBJLEN": lambda args: [((args[1],), 1, 1)],
        b"JSON.TYPE": lambda args: [((args[1],), 1, 1)],
        b"JSON.RESP": lambda args: [((args[1],), 1, 1)],
        b"JSON.TOGGLE": lambda args: [((args[1],), 1, 1)],
        b"JSON.CLEAR": lambda args: [((args[1],), 1, 1)],
        b"JSON.NUMMULTBY": lambda args: [((args[1],), 1, 1)],
        b"JSON.MERGE": lambda args: [((args[1],), 1, 1)],
        b"JSON.MGET": lambda args: [(args[1 : len(args) - 1], 1, len(args) - 2)],
        b"JSON.MSET": lambda args: [(args[1::3], 1, len(args) - 1)],
        b"BF.RESERVE": lambda args: [((args[1],), 1, 1)],
        b"BF.ADD": lambda args: [((args[1],), 1, 1)],
        b"BF.MADD": lambda args: [((args[1],), 1, 1)],
        b"BF.INSERT": lambda args: [((args[1],), 1, 1)],
        b"BF.EXISTS": lambda args: [((args[1],), 1, 1)],
        b"BF.MEXISTS": lambda args: [((args[1],), 1, 1)],
        b"BF.SCANDUMP": lambda args: [((args[1],), 1, 1)],
        b"BF.LOADCHUNK": lambda args: [((args[1],), 1, 1)],
        b"BF.INFO": lambda args: [((args[1],), 1, 1)],
        b"BF.CARD": lambda args: [((args[1],), 1, 1)],
        b"CF.RESERVE": lambda args: [((args[1],), 1, 1)],
        b"CF.ADD": lambda args: [((args[1],), 1, 1)],
        b"CF.ADDNX": lambda args: [((args[1],), 1, 1)],
        b"CF.INSERT": lambda args: [((args[1],), 1, 1)],
        b"CF.INSERTNX": lambda args: [((args[1],), 1, 1)],
        b"CF.EXISTS": lambda args: [((args[1],), 1, 1)],
        b"CF.MEXISTS": lambda args: [((args[1],), 1, 1)],
        b"CF.DEL": lambda args: [((args[1],), 1, 1)],
        b"CF.COUNT": lambda args: [((args[1],), 1, 1)],
        b"CF.SCANDUMP": lambda args: [((args[1],), 1, 1)],
        b"CF.LOADCHUNK": lambda args: [((args[1],), 1, 1)],
        b"CF.INFO": lambda args: [((args[1],), 1, 1)],
        b"CMS.INITBYDIM": lambda args: [((args[1],), 1, 1)],
        b"CMS.INITBYPROB": lambda args: [((args[1],), 1, 1)],
        b"CMS.INCRBY": lambda args: [((args[1],), 1, 1)],
        b"CMS.QUERY": lambda args: [((args[1],), 1, 1)],
        b"CMS.INFO": lambda args: [((args[1],), 1, 1)],
        b"CMS.MERGE": lambda args: [((args[1],) + args[3 : 3 + int(args[2])], 1, 3 + int(args[2]))],
        b"TOPK.RESERVE": lambda args: [((args[1],), 1, 1)],
        b"TOPK.ADD": lambda args: [((args[1],), 1, 1)],
        b"TOPK.INCRBY": lambda args: [((args[1],), 1, 1)],
        b"TOPK.QUERY": lambda args: [((args[1],), 1, 1)],
        b"TOPK.LIST": lambda args: [((args[1],), 1, 1)],
        b"TOPK.INFO": lambda args: [((args[1],), 1, 1)],
        b"TOPK.COUNT": lambda args: [((args[1],), 1, 1)],
        b"TDIGEST.CREATE": lambda args: [((args[1],), 1, 1)],
        b"TDIGEST.RESET": lambda args: [((args[1],), 1, 1)],
        b"TDIGEST.ADD": lambda args: [((args[1],), 1, 1)],
        b"TDIGEST.MERGE": lambda args: [
            ((args[1],) + args[3 : 3 + int(args[2])], 1, 3 + int(args[2]))
        ],
        b"TDIGEST.MIN": lambda args: [((args[1],), 1, 1)],
        b"TDIGEST.MAX": lambda args: [((args[1],), 1, 1)],
        b"TDIGEST.QUANTILE": lambda args: [((args[1],), 1, 1)],
        b"TDIGEST.CDF": lambda args: [((args[1],), 1, 1)],
        b"TDIGEST.TRIMMED_MEAN": lambda args: [((args[1],), 1, 1)],
        b"TDIGEST.RANK": lambda args: [((args[1],), 1, 1)],
        b"TDIGEST.REVRANK": lambda args: [((args[1],), 1, 1)],
        b"TDIGEST.BYRANK": lambda args: [((args[1],), 1, 1)],
        b"TDIGEST.BYREVRANK": lambda args: [((args[1],), 1, 1)],
        b"TDIGEST.INFO": lambda args: [((args[1],), 1, 1)],
        b"TS.CREATE": lambda args: [((args[1],), 1, 1)],
        b"TS.CREATERULE": lambda args: [(args[1:3], 1, 3)],
        b"TS.ALTER": lambda args: [((args[1],), 1, 1)],
        b"TS.ADD": lambda args: [((args[1],), 1, 1)],
        b"TS.MADD": lambda args: [(args[1:-1:3], 1, -1)],
        b"TS.INCRBY": lambda args: [((args[1],), 1, 1)],
        b"TS.DECRBY": lambda args: [((args[1],), 1, 1)],
        b"TS.DELETERULE": lambda args: [(args[1:3], 1, 3)],
        b"TS.GET": lambda args: [((args[1],), 1, 1)],
        b"TS.INFO": lambda args: [((args[1],), 1, 1)],
        b"TS.REVRANGE": lambda args: [((args[1],), 1, 1)],
        b"TS.RANGE": lambda args: [((args[1],), 1, 1)],
        b"TS.DEL": lambda args: [((args[1],), 1, 1)],
        b"FT.CREATE": lambda args: [((args[1],), 1, 1)],
        b"FT.INFO": lambda args: [((args[1],), 1, 1)],
        b"FT.EXPLAIN": lambda args: [((args[1],), 1, 1)],
        b"FT.ALTER": lambda args: [((args[1],), 1, 1)],
        b"FT.ALIASADD": lambda args: [((args[1],), 1, 1)],
        b"FT.ALIASUPDATE": lambda args: [((args[1],), 1, 1)],
        b"FT.ALIASDEL": lambda args: [((args[1],), 1, 1)],
        b"FT.TAGVALS": lambda args: [((args[1],), 1, 1)],
        b"FT.CONFIG GET": lambda args: [((args[1],), 1, 1)],
        b"FT.CONFIG SET": lambda args: [((args[1],), 1, 1)],
        b"FT.SEARCH": lambda args: [((args[1],), 1, 1)],
        b"FT.HYBRID": lambda args: [((args[1],), 1, 1)],
        b"FT.AGGREGATE": lambda args: [((args[1],), 1, 1)],
        b"FT.CURSOR GET": lambda args: [((args[1],), 1, 1)],
        b"FT.CURSOR DEL": lambda args: [((args[1],), 1, 1)],
        b"FT.SYNUPDATE": lambda args: [((args[1],), 1, 1)],
        b"FT.SYNDUMP": lambda args: [((args[1],), 1, 1)],
        b"FT.SPELLCHECK": lambda args: [((args[1],), 1, 1)],
        b"FT.DICTADD": lambda args: [((args[1],), 1, 1)],
        b"FT.DICTDEL": lambda args: [((args[1],), 1, 1)],
        b"FT.DICTDUMP": lambda args: [((args[1],), 1, 1)],
        b"FT.DROPINDEX": lambda args: [((args[1],), 1, 1)],
        b"FT.SUGADD": lambda args: [((args[1],), 1, 1)],
        b"FT.SUGDEL": lambda args: [((args[1],), 1, 1)],
        b"FT.SUGGET": lambda args: [((args[1],), 1, 1)],
        b"FT.SUGLEN": lambda args: [((args[1],), 1, 1)],
    }

    @classmethod
    def extract_keys(
        cls, command: bytes, *arguments: RedisValueT, readonly_command: bool = False
    ) -> tuple[tuple[RedisValueT, ...], tuple[int, int]]:
        """
        Returns the keys from the command + arguments
        """
        try:
            complete_args = (command, *arguments)
            if readonly_command and command in cls.READONLY_VARIANTS:
                ranges = cls.READONLY_VARIANTS[command](complete_args)
            else:
                ranges = cls.COMMANDS[command](complete_args)
            keys: list[RedisValueT] = []
            first_key: int | None = None
            last_key: int | None = None
            for args, first, last in ranges:
                keys.extend(args)
                if first is not None:
                    first_key = min(first_key, first) if first_key is not None else first
                if last is not None:
                    last_key = max(last_key, last) if last_key is not None else last
            if first_key is not None and last_key is not None:
                key_range = (first_key - 1, last_key - 1)
            else:
                key_range = (0, 0)
            return tuple(keys), key_range
        except KeyError:
            return ((), (0, 0))

    @classmethod
    def slots_to_keys(
        cls, command: bytes, *arguments: RedisValueT, readonly_command: bool = False
    ) -> dict[int, list[tuple[int, RedisValueT]]]:
        """
        Returns a mapping of slots to tuples of (original_index, key)
        """
        keys = cls.extract_keys(command, *arguments, readonly_command=readonly_command)[0]
        mapping: dict[int, list[tuple[int, RedisValueT]]] = {}
        for idx, key in enumerate(keys):
            mapping.setdefault(hash_slot(b(key)), []).append((idx, key))
        return mapping

    @classmethod
    def affected_slots(
        cls, command: bytes, *arguments: RedisValueT, readonly_command: bool = False
    ) -> set[int]:
        """
        Returns all slots affected by the command
        """
        keys = cls.extract_keys(command, *arguments, readonly_command=readonly_command)[0]
        return set(hash_slot(b(key)) for key in keys)



from __future__ import annotations

import dataclasses
import inspect
import sys
from collections import OrderedDict
from collections.abc import (
    AsyncGenerator,
    AsyncIterator,
    Awaitable,
    Callable,
    Coroutine,
    Generator,
    Hashable,
    Iterable,
    Iterator,
    Mapping,
    MutableMapping,
    MutableSequence,
    MutableSet,
    Sequence,
    Set,
    ValuesView,
)
from types import GenericAlias, ModuleType, UnionType
from typing import (
    TYPE_CHECKING,
    Any,
    AnyStr,
    AsyncContextManager,
    ClassVar,
    Concatenate,
    Final,
    Generic,
    Literal,
    NamedTuple,
    ParamSpec,
    Protocol,
    TypeAlias,
    TypeGuard,
    TypeVar,
    cast,
    get_origin,
    get_type_hints,
    overload,
    runtime_checkable,
)

from beartype import __version__ as beartype_version
from beartype import beartype
from packaging import version

if TYPE_CHECKING:
    infer_hint: Callable[..., Any]
    is_bearable: Callable[[Any, Any], bool]
    is_subhint: Callable[[Any, Any], bool]
else:
    if version.parse(beartype_version) < version.parse("0.22"):
        from beartype.door import infer_hint
    else:
        from beartype.bite import infer_hint

from beartype.door import is_bearable, is_subhint
from typing_extensions import (
    NotRequired,
    Self,
    TypedDict,
    TypeIs,
    Unpack,
)

from coredis.config import Config

RUNTIME_TYPECHECKS = Config.runtime_checks and not TYPE_CHECKING

P = ParamSpec("P")
T_co = TypeVar("T_co", covariant=True)
R = TypeVar("R")


def safe_beartype(func: Callable[P, R]) -> Callable[P, R]:
    return beartype(func)


def add_runtime_checks(func: Callable[P, R]) -> Callable[P, R]:
    if RUNTIME_TYPECHECKS and not TYPE_CHECKING:
        return beartype(func)

    return func


class RedisError(Exception):
    """
    Base exception from which all other exceptions in coredis
    derive from.
    """


class Node(TypedDict):
    """
    Definition of a cluster node
    """

    host: str
    port: int


class RedisCommandP(Protocol):
    """
    Protocol of a redis command with all associated arguments
    converted into the shape expected by the redis server.
    Used by :meth:`~coredis.Redis.execute_command`
    """

    #: The name of the redis command
    name: bytes
    #: All arguments to be passed to the command
    arguments: tuple[RedisValueT, ...]


@dataclasses.dataclass
class RedisCommand:
    """
    Convenience data class that conforms to :class:`~coredis.typing.RedisCommandP`
    """

    #: The name of the redis command
    name: bytes
    #: All arguments to be passed to the command
    arguments: tuple[RedisValueT, ...]


class ExecutionParameters(TypedDict):
    """
    Extra parameters that can be passed to :meth:`~coredis.Redis.execute_command`
    """

    #: Whether to decode the response
    #: (ignoring the value of :paramref:`~coredis.Redis.decode_responses`)
    decode: NotRequired[bool]
    slot_arguments_range: NotRequired[tuple[int, int]]


#: Represents the acceptable types of a redis key
KeyT: TypeAlias = str | bytes


class Serializable(Generic[R]):
    """
    Wrapper to be used to pass arbitrary types to redis commands
    to be eventually serialized by :class:`coredis.typing.TypeAdapter.serialize`

    Wrapping a value in :class:`Serializable` will pass type checking
    wherever a method expects a :class:`coredis.typing.ValueT` - however
    it will still fail if there is no serializer registered through the instance
    of :class:`coredis.typing.TypeAdapter` that is associated with the client.

    For example::

      class MyThing:
          ...

      async with coredis.Redis() as client:
          # This will pass type checking but will fail with an :exc:`LookupError`
          # at runtime
          await client.set("fubar", coredis.typing.Serializable(MyThing()))

      # however, if a serializer is registered, the above would succeed
      @client.type_adapter.serializer
      def _(value: MyThing) -> str:
         ... # some way to convert it to a string
    """

    def __init__(self, value: R) -> None:
        self.value = value


AdaptableType = type | UnionType | GenericAlias


class TypeAdapter:
    """
    Used by the coredis clients :class:`~coredis.Redis` and :class:`~coredis.RedisCluster`
    through :paramref:`~coredis.Redis.type_adapter` for adapting complex types that require
    custom serialization/deserialization with redis commands.

    For example to use Decimal types with some common redis operations::

      from decimal import Decimal
      from typing import Any, Mapping, Iterable
      from coredis import Redis
      from coredis.typing import TypeAdapter, Serializable

      adapter = TypeAdapter()

      @adapter.serializer
      def decimal_to_str(value: Decimal) -> str:
          return str(value)

      @adapter.deserializer
      def value_to_decimal(value: str|bytes) -> Decimal:
          return Decimal(value.decode("utf-8") if isinstance(value, bytes) else value)

      @adapter.deserializer
      def list_to_decimal_list(items: Iterable[str|bytes]) -> list[Decimal]:
          return [value_to_decimal(value) for value in items]

      @adapter.deserializer
      def mapping_to_decimal_mapping(mapping: Mapping[str|bytes, str|bytes]) -> dict[str|bytes, Decimal]:
          return {key: value_to_decimal(value) for key, value in mapping.items()}

      async with coredis.Redis(type_adapter=adapter, decode_responses=True) as client:
          await client.set("key", Serializable(Decimal(1.5)))
          await client.lpush("list", [Serializable(Decimal(1.5))])
          await client.hset("dict", {"first": Serializable(Decimal(1.5))})
          assert Decimal(1.5) == await client.get("key").transform(Decimal)
          assert [Decimal(1.5)] == await client.lrange("list", 0, 0).transform(list[Decimal])
          assert {"first": Decimal(1.5)} == await client.hgetall("dict").transform(dict[str, Decimal])
    """

    def __init__(
        self,
    ) -> None:
        self.__serializers: dict[
            AdaptableType,
            tuple[Callable[[Any], RedisValueT], int],
        ] = {}
        self.__deserializers: dict[
            AdaptableType,
            dict[AdaptableType, tuple[Callable[..., Any], int]],
        ] = {}
        self.__deserializer_cache: dict[
            tuple[AdaptableType, AdaptableType | GenericAlias],
            Callable[..., Any],
        ] = {}
        self.__serializer_cache: dict[AdaptableType, Callable[[Any], RedisValueT]] = {}

    @classmethod
    def format_type(cls, type_like: AdaptableType) -> str:
        if get_origin(type_like):
            return str(type_like)
        else:
            return getattr(type_like, "__name__", str(type_like))

    def register(
        self,
        type: type[R] | UnionType,
        serializer: Callable[[R], RedisValueT],
        deserializer: Callable[[Any], R],
        deserializable_type: type = object,
    ) -> None:
        """
        Register both a serializer and a deserializer for :paramref:`type`

        :param type: The type that should be serialized/deserialized
        :param serializer: a function that receives an instance of :paramref:`type`
         and returns a value of type :data:`coredis.typing.RedisValueT`
        :param deserializer: a function that accepts the return types from
         the redis commands that are expected to be used when deserializing
         to :paramref:`type`.
        :param deserializable_type: the types of values :paramref:`deserializer` should
         be considered for
        """
        self.register_serializer(type, serializer)
        self.register_deserializer(type, deserializer, deserializable_type)

    def register_serializer(
        self,
        serializable_type: type[R] | UnionType,
        serializer: Callable[[R], RedisValueT],
    ) -> None:
        """
        Register a serializer for :paramref:`type`

        :param type: The type that will be serialized
        :param serializer: a function that receives an instance of :paramref:`type`
         and returns a value of type :data:`coredis.typing.RedisValueT`
        """
        self.__serializers.setdefault(serializable_type, (serializer, 0))
        self.__serializer_cache.clear()

    def register_deserializer(
        self,
        deserialized_type: type[R] | UnionType,
        deserializer: Callable[[Any], R],
        deserializable_type: AdaptableType = object,
    ) -> None:
        """
        Register a deserializer for :paramref:`type` and automatically register
        deserializers for common collection types that use this type.

        :param type: The type that should be deserialized
        :param deserializer: a function that accepts the return types from
         the redis commands that are expected to be used when deserializing
         to :paramref:`type`.
        :param deserializable_type: the types of values :paramref:`deserializer` should
         be considered for
        """

        def register_collection_deserializer(
            collection_type: AdaptableType,
            deserializable_type: AdaptableType,
            deserializer: Callable[[Any], Any],
        ) -> None:
            self.__deserializers.setdefault(collection_type, {}).setdefault(
                deserializable_type,
                (deserializer, -1),
            )

        # Register the base deserializer
        self.__deserializers.setdefault(deserialized_type, {})[deserializable_type or object] = (
            deserializer,
            0,
        )

        # Register collection deserializers
        register_collection_deserializer(
            GenericAlias(list, (deserialized_type,)),
            GenericAlias(Iterable, deserializable_type),
            lambda v: [deserializer(item) for item in v],
        )
        register_collection_deserializer(
            GenericAlias(set, (deserialized_type,)),
            GenericAlias(Iterable, deserializable_type),
            lambda v: {deserializer(item) for item in v},
        )
        register_collection_deserializer(
            GenericAlias(tuple, (deserialized_type, ...)),
            GenericAlias(Iterable, deserializable_type),
            lambda v: tuple([deserializer(item) for item in v]),
        )

        # Register dictionary deserializers for existing types
        for t in list(self.__deserializers):
            if t != deserialized_type:
                for rt in list(self.__deserializers[t]):
                    _deserializer, priority = self.__deserializers[t][rt]
                    if priority >= 0:
                        register_collection_deserializer(
                            GenericAlias(dict, (t, deserialized_type)),
                            GenericAlias(Mapping, (rt, deserializable_type)),
                            lambda m, key_deserializer=_deserializer: {  # type: ignore
                                key_deserializer(k): deserializer(v) for k, v in m.items()
                            },
                        )
                        register_collection_deserializer(
                            GenericAlias(dict, (deserialized_type, t)),
                            GenericAlias(Mapping, (deserializable_type, rt)),
                            lambda m, value_deserializer=_deserializer: {  # type: ignore
                                deserializer(k): value_deserializer(v) for k, v in m.items()
                            },
                        )

        # Register dictionary deserializers for primitive types
        for t in {bytes, str}:
            register_collection_deserializer(
                GenericAlias(dict, (t, deserialized_type)),
                GenericAlias(Mapping, (t, deserializable_type)),
                lambda v: {k: deserializer(v) for k, v in v.items()},
            )
            register_collection_deserializer(
                GenericAlias(dict, (deserialized_type, t)),
                GenericAlias(Mapping, (deserializable_type, t)),
                lambda v: {deserializer(k): v for k, v in v.items()},
            )

        self.__deserializer_cache.clear()

    def serializer(self, func: Callable[[R], RedisValueT]) -> Callable[[R], RedisValueT]:
        """
        Decorator for registering a serializer

        :param func: A serialization function that accepts an instance of
         type `R` and returns one of the types defined by :data:`coredis.typing.RedisValueT`
         The acceptable  serializable types are inferred
         from the annotations in the function signature.

        :raises ValueError: when the appropriate serializable type cannot be
         inferred.
        """
        if (parameters := list(inspect.signature(func).parameters.keys())) and (
            input_hint := get_type_hints(func).get(parameters[0])
        ):
            self.register_serializer(input_hint, func)
            return func
        else:
            raise ValueError(
                "Unable to infer custom input type from decorated function. Check type annotations."
            )

    def deserializer(self, func: Callable[[Any], R]) -> Callable[[Any], R]:
        """
        Decorator for registering a deserializer

        :param func: A deserialization function that returns an instance of
         type `R` that can be used with :meth:`deserialize`. The acceptable
         deserializable types and the expected deserialized type are inferred
         from the annotations in the function signature.

        :raises ValueError: when the appropriate input/output types cannot be
         inferred.
        """
        if (
            (parameters := list(inspect.signature(func).parameters.keys()))
            and (input_hint := get_type_hints(func).get(parameters[0]))
        ) and (response_type := get_type_hints(func).get("return")):
            self.register_deserializer(response_type, func, input_hint)
            return func
        else:
            raise ValueError(
                "Unable to infer response type from decorated function. Check annotations."
            )

    def serialize(self, value: Serializable[R]) -> RedisValueT:
        """
        Serializes :paramref:`value` into one of the types represented by
        :data:`~coredis.typing.RedisValueT` using a serializer registered
        via :meth:`register_serializer` or decorated by :meth:`serializer`.

        :param: a value wrapped in :class:`coredis.typing.Serializable`
        """
        value_type = cast(AdaptableType, infer_hint(value.value))
        if not (transform_function := self.__serializer_cache.get(value_type, None)):
            candidate: tuple[AdaptableType, Callable[[R], RedisValueT] | None] = (object, None)

            for t in self.__serializers:
                if is_bearable(value.value, t):
                    if not candidate[1] or is_subhint(t, candidate[0]):
                        candidate = (t, self.__serializers[t][0])
            if candidate[1]:
                transform_function = candidate[1]
                self.__serializer_cache[value_type] = transform_function
        if not transform_function:
            raise LookupError(
                f"No registered serializer to serialize {self.format_type(value_type)}"
            )
        return transform_function(value.value)

    def deserialize(self, value: Any, return_type: type[R]) -> R:
        """
        Deserializes :paramref:`value` into an instance of :paramref:`return_type`
        using a deserializer registered via :meth:`register_deserializer` or decorated
        by :meth:`deserializer`.

        :param value: the value to be deserialized (typically something returned by one of
         the redis commands)
        :param return_type: The type to deserialize to
        """
        value_type = cast(AdaptableType, infer_hint(value))
        if not (deserializer := self.__deserializer_cache.get((value_type, return_type), None)):
            if exact_match := self.__deserializers.get(return_type, {}).get(value_type, None):
                deserializer = exact_match[0]
            else:
                candidate: tuple[AdaptableType, AdaptableType, Callable[[Any], R] | None, int] = (
                    object,
                    object,
                    None,
                    -100,
                )
                for registered_type, transforms in self.__deserializers.items():
                    if is_subhint(return_type, registered_type):
                        for expected_value_type in transforms:
                            if (
                                is_bearable(value, expected_value_type)
                                and is_subhint(registered_type, candidate[0])
                                and is_subhint(expected_value_type, candidate[1])
                                and transforms[expected_value_type][1] >= candidate[3]
                            ):
                                candidate = (
                                    registered_type,
                                    expected_value_type,
                                    transforms[expected_value_type][0],
                                    transforms[expected_value_type][1],
                                )
                deserializer = candidate[2]
        if deserializer:
            deserialized = deserializer(value)
            if RUNTIME_TYPECHECKS and not is_subhint(
                transformed_type := cast(type, infer_hint(deserialized)), return_type
            ):
                raise TypeError(
                    f"Invalid deserializer. Requested {self.format_type(return_type)} but deserializer returned {self.format_type(transformed_type)}"
                )
            self.__deserializer_cache[(value_type, return_type)] = deserializer
            return deserialized
        elif is_subhint(value_type, return_type):
            return cast(R, value)
        else:
            raise LookupError(
                f"No registered deserializer to convert {self.format_type(value_type)} to {self.format_type(return_type)}"
            )


#: Represents the different python primitives that are accepted
#: as input parameters for commands that can be used with loosely
#: defined types. These will eventually be serialized before being
#: sent to redis.
#:
#: Additionally any object wrapped in a :class:`Serializable` will be
#: accepted and will be serialized using an appropriate type adapter
#: registered with the client. See :ref:`api/typing:custom types` for more details.
ValueT: TypeAlias = str | bytes | int | float | Serializable[Any]
#: The canonical type used for input parameters that represent "strings"
#: that are transmitted to redis.
StringT: TypeAlias = str | bytes

#: Used for dictionary keys for all commands that accept :class:`~typing.Mapping`
#: as a parameter to allow any :class:`~coredis.typing.ValueT` as a key
MappingKeyT = TypeVar("MappingKeyT", bound=ValueT)

CommandArgList = list[ValueT]

#: Primitive types that we can expect to be sent to redis with
#: simple serialization. The internals of coredis
#: pass around arguments to redis commands as this type.
RedisValueT: TypeAlias = str | bytes | int | float

#: Restricted union of container types accepted as arguments to apis
#: that accept a variable number values for an argument (such as keys, values).
#: This is used instead of :class:`typing.Iterable` as the latter allows
#: :class:`str` to be passed in as valid values for :class:`Iterable[str]` or :class:`bytes`
#: to be passed in as a valid value for :class:`Iterable[bytes]` which is never the actual
#: expectation in the scope of coredis.
#: For example::
#:
#:     def length(values: Parameters[ValueT]) -> int:
#:         return len(list(values))
#:
#:     length(["1", 2, 3, 4])      # valid
#:     length({"1", 2, 3, 4})      # valid
#:     length(("1", 2, 3, 4))      # valid
#:     length({"1": 2}.keys())     # valid
#:     length({"1": 2}.values())   # valid
#:     length(map(str, range(10))) # valid
#:     length({"1": 2})            # invalid
#:     length("123")               # invalid
#:     length(b"123")              # invalid


@runtime_checkable
class SequenceNotString(Protocol[T_co]):
    """
    Allow sequences *except* str | bytes.

    Use this to disallow passing a single str | bytes instance where a sequence of
    strings is expected.
    """

    def __contains__(self, value: object, /) -> bool: ...

    @overload
    def __getitem__(self, index: int, /) -> T_co: ...

    @overload
    def __getitem__(self, index: slice, /) -> Sequence[T_co]: ...

    def __len__(self) -> int: ...

    def __iter__(self) -> Iterator[T_co]: ...

    def __reversed__(self, /) -> Iterator[T_co]: ...


Parameters = (
    SequenceNotString[T_co] | Set[T_co] | tuple[T_co, ...] | ValuesView[T_co] | Iterator[T_co]
)

if sys.version_info >= (3, 12):
    from ._py_312_typing import JsonType, ResponsePrimitive, ResponseType
else:
    from ._py_311_typing import JsonType, ResponsePrimitive, ResponseType


__all__ = [
    "Serializable",
    "AnyStr",
    "AsyncContextManager",
    "AsyncIterator",
    "AsyncGenerator",
    "Awaitable",
    "Callable",
    "ClassVar",
    "CommandArgList",
    "Concatenate",
    "Coroutine",
    "Final",
    "Generic",
    "Generator",
    "Hashable",
    "Iterable",
    "Iterator",
    "JsonType",
    "KeyT",
    "Literal",
    "MappingKeyT",
    "Mapping",
    "ModuleType",
    "MutableMapping",
    "MutableSet",
    "MutableSequence",
    "NamedTuple",
    "Node",
    "NotRequired",
    "OrderedDict",
    "Parameters",
    "ParamSpec",
    "Protocol",
    "RedisCommand",
    "RedisCommandP",
    "ExecutionParameters",
    "ResponsePrimitive",
    "ResponseType",
    "runtime_checkable",
    "Self",
    "Sequence",
    "StringT",
    "TypeGuard",
    "TypeIs",
    "TypedDict",
    "TypeVar",
    "Unpack",
    "ValueT",
    "RedisValueT",
    "ValuesView",
    "TYPE_CHECKING",
    "TypeAdapter",
    "ValueT",
    "RUNTIME_TYPECHECKS",
]



from __future__ import annotations

import dataclasses
from _weakref import ProxyType, proxy
from typing import TYPE_CHECKING, Any

from anyio import Event, get_cancelled_exc_class, move_on_after

from coredis._utils import nativestr
from coredis.typing import Generator, RedisValueT, ResponseType

if TYPE_CHECKING:
    from ._base import BaseConnection


@dataclasses.dataclass(slots=True)
class Request:
    connection: dataclasses.InitVar[BaseConnection]
    command: bytes
    args: tuple[RedisValueT, ...]
    decode: bool
    encoding: str | None = None
    raise_exceptions: bool = True
    response_timeout: float | None = None
    disconnect_on_cancellation: bool = False
    expects_response: bool = True

    _connection: ProxyType[BaseConnection] = dataclasses.field(init=False)
    _event: Event = dataclasses.field(init=False, default_factory=Event)
    _exc: BaseException | None = dataclasses.field(init=False, default=None)
    _result: ResponseType | None = dataclasses.field(init=False, default=None)

    def __post_init__(self, connection: BaseConnection) -> None:
        self._connection = proxy(connection)
        self._connection.statistics.request_created()
        if not self.expects_response:
            self.resolve(None)

    def __await__(self) -> Generator[Any, None, ResponseType]:
        return self.get_result().__await__()

    def resolve(self, response: ResponseType) -> None:
        self._connection.statistics.request_resolved()
        self._result = response
        self._event.set()

    def fail(self, error: BaseException) -> None:
        if not self._event.is_set():
            self._connection.statistics.request_resolved()
            self._exc = error
            self._event.set()

    async def get_result(self) -> ResponseType:
        if not self._event.is_set():
            try:
                with move_on_after(self.response_timeout) as scope:
                    await self._event.wait()
                if scope.cancelled_caught and not self._event.is_set():
                    reason = (
                        f"{nativestr(self.command)} timed out after {self.response_timeout} seconds"
                    )
                    self._exc = TimeoutError(reason)
                    self._handle_response_cancellation(reason)
            except get_cancelled_exc_class():
                self._handle_response_cancellation(f"{nativestr(self.command)} was cancelled")
                raise
        return self._result_or_exc()

    def _handle_response_cancellation(self, reason: str) -> None:
        if self._connection and self.disconnect_on_cancellation:
            self._connection.invalidate(reason)

    def _result_or_exc(self) -> ResponseType:
        if self._exc is not None:
            if self.raise_exceptions:
                raise self._exc
            return self._exc  # type: ignore
        return self._result



from __future__ import annotations

from coredis.response._callbacks import ResponseCallback
from coredis.typing import (
    AnyStr,
    ResponsePrimitive,
)


class ACLLogCallback(
    ResponseCallback[
        list[dict[AnyStr, ResponsePrimitive] | None],
        tuple[dict[AnyStr, ResponsePrimitive] | None, ...],
    ]
):
    def transform(
        self,
        response: list[dict[AnyStr, ResponsePrimitive] | None],
    ) -> tuple[dict[AnyStr, ResponsePrimitive] | None, ...]:
        return tuple(response)



from __future__ import annotations

import enum

from coredis.typing import StringT


@enum.unique
class CaseAndEncodingInsensitiveEnum(bytes, enum.Enum):
    __decoded: set[StringT]

    @property
    def variants(self) -> set[StringT]:
        if not hasattr(self, "__decoded"):
            decoded = str(self)
            self.__decoded = {
                self.value.lower(),  # type: ignore
                self.value,  # type: ignore
                decoded.lower(),
                decoded.upper(),
            }
        return self.__decoded

    def __eq__(self, other: object) -> bool:
        """
        Since redis tokens are case insensitive allow mixed case
        Additionally allow strings to be passed in instead of
        bytes.
        """

        if other:
            if isinstance(other, self.__class__):
                return bool(self.value == other.value)
            else:
                return other in self.variants
        return False

    def __str__(self) -> str:
        return self.decode("latin-1")

    def __hash__(self) -> int:
        return hash(self.value)
